#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
import cvxopt as opt
from ..defs import *
from ..scalars import cvxpy_obj,cvxpy_scalar_var
from ..utils import vstack,hstack
from ..interface import belongs
from ..interface import variable
from ..interface import greater_equals
from ..constraints import cvxpy_list
from .semidefinite_cone import semidefinite_cone
from functools import reduce

# Class definition
class cvxpy_geo_mean_cone(object):
    r"""
    | :math:`\{(x,y) \in \mathbb{R}_+^n \\times \mathbb{R} \ |
      \ \\big( \prod_{i=1}^{n}x_i \\big)^{1/n} \ge y \}`.
    """

    # Method: __init__
    def __init__(self):
        """
        Class constructor.
        """

        self.type = SET
        self.name = 'geo_mean_cone'
        self.expansion_type = PM

    # Method: valid_shape
    def valid_shape(self,shape):

        return shape[0]>=2 and shape[1] == 1

    # Method: __str__
    def __str__(self):

        return self.name

    # Method: _pm_expand
    def _pm_expand(self,constr):

        # Get objects
        v = constr.left
        v = vstack([v[i,0] for i in range(0,v.shape[0])])
        n = v.shape[0]-1
        x = vstack([v[0:n,0]])
        y = v[n,0]
        z = variable()

        # One element
        if n==1:
            return cvxpy_list([greater_equals(x,0),
                               greater_equals(x,y)])

        # Get power of 2 size
        m = 0
        while np.log2(n+m) % 1 != 0:
            m += 1

        # Copy elements of x on a list and restrict them
        constr_list = []
        el_list = []
        for i in range(0,n,1):
            el_list+=[x[i,0]]
            if (not np.isscalar(x[i,0]) and
                type(x[i,0]) is not cvxpy_obj):
                constr_list += [greater_equals(x[i,0],0)]

        # Construct expansion
        for i in range(0,m,1):
            el_list += [z]
        while len(el_list) > 2:
            new_list = []
            for i in range(0,int(len(el_list)/2)):
                x1 = el_list[2*i]
                x2 = el_list[2*i+1]
                w = variable()
                constr_list += [belongs(vstack((hstack((x1,w)),
                                                hstack((w,x2)))),
                                        semidefinite_cone)]
                new_list += [w]
            el_list = new_list
        x1 = el_list[0]
        x2 = el_list[1]
        constr_list += [belongs(vstack((hstack((x1,z)),
                                        hstack((z,x2)))),
                                semidefinite_cone)]
        constr_list += [greater_equals(z,0),greater_equals(z,y)]
        return cvxpy_list(constr_list)

    # Method: _construct
    def _construct(self,el,mp,p):
        n = el.shape[0]-1
        u = vstack([el[0:n,0]])
        v = el[n,0]

        # Prepare u
        for i in range(0,n,1):
            if type(u[i,0]) is cvxpy_obj:
                u[i,0] = u[i,0].value

        # Prepare v
        if type(v) is cvxpy_obj:
            v = v.value

        # f
        def f(x):
            vals = []
            for i in range(0,n,1):
                if type(u[i,0]) is cvxpy_scalar_var:
                    vals += [x[mp[u[i,0]]]]
                else:
                    vals += [u[i,0]]
            if type(v) is cvxpy_scalar_var:
                y = x[mp[v]]
            else:
                y = v
            return y - reduce(lambda w,z:w*z,vals,1.)**(1./(n*1.))

        # grad f
        def grad_f(x):
            g = opt.spmatrix(0.0,[],[],(p,1))
            if type(v) is cvxpy_scalar_var:
                fval = -f(x)+x[mp[v]]
            else:
                fval = -f(x)+v
            for i in range(0,n,1):
                if type(u[i,0]) is cvxpy_scalar_var:
                    xi = x[mp[u[i,0]]]
                    g[mp[u[i,0]]] = -fval/(n*xi*1.)
            if type(v) is cvxpy_scalar_var:
                g[mp[v]] = 1.
            return g

        # hess f
        def hess_f(x):
            h = opt.spmatrix(0.0,[],[],(p,p))
            if type(v) is cvxpy_scalar_var:
                fval = -f(x)+x[mp[v]]
            else:
                fval = -f(x)+v
            for i in range(0,n,1):
                for j in range(0,n,1):
                    if (type(u[i,0]) is cvxpy_scalar_var and
                        type(u[j,0]) is cvxpy_scalar_var):
                        xi = x[mp[u[i,0]]]
                        xj = x[mp[u[j,0]]]
                        if i == j:
                            w = (n-1.)*fval/(n*n*1.*xi*xi)
                            h[mp[u[i,0]],mp[u[i,0]]] = w
                        else:
                            w = -fval/(1.*n*n*xi*xj)
                            h[mp[u[i,0]],mp[u[j,0]]] = w
                            h[mp[u[j,0]],mp[u[i,0]]] = w
            return h

        # ind f
        def ind_f(x):
            for i in range(0,n,1):
                if type(u[i,0]) is cvxpy_scalar_var:
                    if x[mp[u[i,0]]] <= 0:
                        return False
                else:
                    if u[i,0] <= 0:
                        return False
            return True

        # Return functions
        return f, grad_f, hess_f, ind_f

# Create instance
geo_mean_cone = cvxpy_geo_mean_cone()



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from scipy import signal
from pydsm.NTFdesign import ntf_fir_weighting, quantization_noise_gain
from pydsm.delsig import evalTF
import pytest

try:
    import pytest_benchmark
    BENCHMARK_AVAILABLE = True
except ImportError:
    BENCHMARK_AVAILABLE = False

@pytest.mark.skipif(not BENCHMARK_AVAILABLE,
                    reason="pytest-benchmark is not installed")
@pytest.mark.benchmark(group="ntf_design-fir-weighting")
class Benchmark_ntf_fir_weighting(object):

    @classmethod
    def setup_class(cls):
        # signal and filter specification
        fsig = 1000.
        B = 400.
        OSR = 64
        fphi = B*OSR*2
        # Lee constraint
        cls.H_inf = 1.5
        # FIR Order
        cls.order = 25
        w0 = 2*fsig/fphi
        B0 = 2*B/fphi
        w1 = (np.sqrt(B0**2+4*w0**2)-B0)/2
        w2 = (np.sqrt(B0**2+4*w0**2)+B0)/2
        cls.hz = signal.butter(4, [w1, w2], 'bandpass', output='zpk')
        cls.ff = np.linspace(0, 0.5, 1024)

    @classmethod
    def teardown_class(cls):
        pass

    def benchmark_ntf_fir_weighting_cvxpy_old(self, benchmark):
        try:
            from pydsm import cvxpy_tdr     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy_old' not installed")
        ntf = benchmark(ntf_fir_weighting,
                        self.order, self.hz, self.H_inf,
                        modeler='cvxpy_old', show_progress=False)
        mf = quantization_noise_gain(ntf, self.hz)
        vv = np.abs(evalTF(ntf, np.exp(2j*np.pi*self.ff)))
        pe = np.max(vv)-self.H_inf
        benchmark.extra_info['Constraints delta'] = pe
        benchmark.extra_info['Quantization noise'] = mf

    def benchmark_ntf_fir_weighting_cvxpy_cvxopt(self, benchmark):
        try:
            import cvxpy     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy' not installed")
        ntf = benchmark(ntf_fir_weighting,
                        self.order, self.hz, self.H_inf,
                        modeler='cvxpy', show_progress=False)
        mf = quantization_noise_gain(ntf, self.hz)
        vv = np.abs(evalTF(ntf, np.exp(2j*np.pi*self.ff)))
        pe = np.max(vv)-self.H_inf
        benchmark.extra_info['Constraints delta'] = pe
        benchmark.extra_info['Quantization noise'] = mf

    def benchmark_ntf_fir_weighting_cvxpy_cvxopt_kkt(self, benchmark):
        try:
            import cvxpy     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy' not installed")
        ntf = benchmark(ntf_fir_weighting,
                        self.order, self.hz, self.H_inf,
                        modeler='cvxpy', show_progress=False,
                        cvxpy_opts={'override_kktsolver': False})
        mf = quantization_noise_gain(ntf, self.hz)
        vv = np.abs(evalTF(ntf, np.exp(2j*np.pi*self.ff)))
        pe = np.max(vv)-self.H_inf
        benchmark.extra_info['Constraints delta'] = pe
        benchmark.extra_info['Quantization noise'] = mf

    def benchmark_ntf_fir_weighting_cvxpy_scs(self, benchmark):
        try:
            import cvxpy     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy' not installed")
        ntf = benchmark(ntf_fir_weighting,
                        self.order, self.hz, self.H_inf,
                        modeler='cvxpy', show_progress=False,
                        cvxpy_opts={'solver': 'scs'})
        mf = quantization_noise_gain(ntf, self.hz)
        vv = np.abs(evalTF(ntf, np.exp(2j*np.pi*self.ff)))
        pe = np.max(vv)-self.H_inf
        benchmark.extra_info['Constraints delta'] = pe
        benchmark.extra_info['Quantization noise'] = mf

    def benchmark_ntf_fir_weighting_picos(self, benchmark):
        try:
            import picos     # analysis:ignore
        except:
            pytest.skip("Modeler 'picos' not installed")
        ntf = benchmark(ntf_fir_weighting,
                        self.order, self.hz, self.H_inf,
                        modeler='picos', show_progress=False)
        mf = quantization_noise_gain(ntf, self.hz)
        vv = np.abs(evalTF(ntf, np.exp(2j*np.pi*self.ff)))
        pe = np.max(vv)-self.H_inf
        benchmark.extra_info['Constraints delta'] = pe
        benchmark.extra_info['Quantization noise'] = mf



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import division

import numpy as np
from warnings import warn
from ..exceptions import PyDsmApproximationWarning
from ._tf import evalTF
from ..relab import cplxpair
from ._ds import ds_optzeros

import sys
if sys.version_info < (3,):
    range = xrange


def synthesizeNTF0(order, osr, opt, H_inf, f0):
    # Determine the zeros.
    if f0 != 0:
        # Bandpass design-- halve the order temporarily.
        order = order//2
        dw = np.pi/(2*osr)
    else:
        dw = np.pi/osr

    if opt.ndim == 0:
        # opt is a number
        if opt == 0:
            z = np.zeros(order)
        else:
            z = dw*ds_optzeros(order, opt)
        if z.size == 0:
            raise RuntimeError('Cannot synthesize NTF zeros')
        if f0 != 0:
            # Bandpass design-- shift and replicate the zeros.
            order = order*2
            z = z + 2*np.pi*f0
            z = np.vstack((z, -z)).transpose().flatten()
        z = np.exp(1j*z)
    else:
        z = opt

    p = np.zeros(order)
    k = 1
    itn_limit = 100
    fprev = 0

    if f0 == 0:
        # Lowpass design
        HinfLimit = 2**order
        # !!! The limit is actually lower for opt=1 and low OSR
        if H_inf >= HinfLimit:
            warn('Unable to achieve specified H_inf, '
                 'setting all NTF poles to zero',
                 PyDsmApproximationWarning)
            p = np.zeros(order)
        else:
            x = 0.3**(order-1)   # starting guess
            for itn in range(1, itn_limit+1):
                me2 = -0.5*(x**(2./order))
                w = (2*np.arange(1, order+1)+1)*np.pi/order
                mb2 = 1+me2*np.exp(1j*w)
                p = mb2 - np.sqrt(mb2**2-1)
                # Reflect poles to be inside the unit circle
                out = abs(p) > 1
                p[out] = 1/p[out]
                # The following is not exactly what delsig does.
                # We do not have an identical cplxpair
                p = cplxpair(p)
                f = np.real(evalTF((z, p, k), -1))-H_inf
                if itn == 1:
                    delta_x = -f/100
                else:
                    delta_x = -f*delta_x/(f-fprev)
                xplus = x+delta_x
                if xplus > 0:
                    x = xplus
                else:
                    x = x*0.1
                fprev = f
                if abs(f) < 1e-10 or abs(delta_x) < 1e-10:
                    break
                if x > 1e6:
                    warn('Unable to achieve specified Hinf, '
                         'setting all NTF poles to zero',
                         PyDsmApproximationWarning)
                    p = np.zeros(order)
                    break
                if itn == itn_limit:
                    warn('Iteration limit exceeded',
                         PyDsmApproximationWarning)
    else:
        # Bandpass design
        x = 0.3**(order//2-1)   # starting guess (not very good for f0~0)
        if f0 > 0.25:
            z_inf = 1.
        else:
            z_inf = -1.
        c2pif0 = np.cos(2*np.pi*f0)
        for itn in range(1, itn_limit+1):
            e2 = 0.5*x**(2./order)
            w = (2*np.arange(order)+1)*np.pi/order
            mb2 = c2pif0 + e2*np.exp(1j*w)
            p = mb2 - np.sqrt(mb2**2-1)
            # Reflect poles to be inside the unit circle
            out = abs(p) > 1
            p[out] = 1/p[out]
            # The following is not exactly what delsig does.
            p = cplxpair(p)
            f = np.real(evalTF((z, p, k), z_inf))-H_inf
            if itn == 1:
                delta_x = -f/100
            else:
                delta_x = -f*delta_x/(f-fprev)
            xplus = x+delta_x
            if xplus > 0:
                x = xplus
            else:
                x = x*0.1
            fprev = f
            if abs(f) < 1e-10 or abs(delta_x) < 1e-10:
                break
            if x > 1e6:
                warn('Unable to achieve specified Hinf, '
                     'Setting all NTF poles to zero',
                     PyDsmApproximationWarning)
                p = np.zeros(order)
                break
            if itn == itn_limit:
                warn('Iteration limit exceeded', PyDsmApproximationWarning)

    z = cplxpair(z)
    return z, p, k



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Collection of DELSIG style plotting routines
============================================
"""

import numpy as np
import matplotlib.pyplot as plt
from ..relab import cplxpair
from scipy.signal import tf2zpk

__all__ = ["plotPZ"]


def plotPZ(H, color='b', markersize=5, showlist=False):
    """
    Plots the poles and zeros of a transfer function.

    Parameters
    ----------
    H : tuple
        transfer function in pzk or nd form
    showlist : bool
        if showlist is true, a list of the poles and zeros is superimposed
        onto the plot.

    Other Parameters
    ----------------
    color : string or list of strings, optional
        color or colors to plot the poles and the zeros (defaults to 'b'
        meaning black)
    markersize : real, optional
        size of the markers used to represent the poles and the zeros
        (defaults to 5)

    Notes
    -----
    See `matplotlib` for info about color codes.

    """

    pole_fmt = {'marker': 'x', 'markersize': markersize}
    zero_fmt = {'marker': 'o', 'markersize': markersize}

    if isinstance(color, list):
        pole_fmt['color'] = color[0]
        zero_fmt['color'] = color[1]
    else:
        pole_fmt['color'] = color
        zero_fmt['color'] = color

    if len(H) == 2:
        H = tf2zpk(*H)
    z = cplxpair(H[0])
    p = cplxpair(H[1])

    plt.grid(True)

    # Plot x and o for poles and zeros, respectively
    plt.plot(p.real, p.imag, linestyle='None', **pole_fmt)
    if len(z) > 0:
        plt.plot(z.real, z.imag, linestyle='None', **zero_fmt)

    # Draw unit circle, real axis and imag axis
    circle = np.exp(2j*np.pi*np.linspace(0, 1, 100))
    plt.plot(circle.real, circle.imag, 'k')
    plt.axis('equal')
    limits = plt.axis()
    plt.plot([0, 0], limits[1:3], 'k:')
    plt.plot(limits[0:2], [0, 0], 'k:')

    if showlist:
        # List the poles and zeros
        pp = p[p.imag >= 0]
        y = 0.05*(len(pp)+1)
        str_p = 'Poles:'
        plt.text(-0.9, y, str_p,
                 horizontalalignment='left',
                 verticalalignment='center')
        y = y - 0.1
        for i in range(len(pp)):
            if pp[i].imag == 0:
                str_p = '$%+.4f$' % pp[i].real
            else:
                str_p = r'$%+.4f\pm \mathrm{j}%.4f$' % (pp[i].real, pp[i].imag)
            plt.text(-0.9, y, str_p,
                     horizontalalignment='left',
                     verticalalignment='center')
            y = y - 0.1
        if len(z) > 0:
            zz = z[z.imag >= 0]
            y = 0.05*(len(zz)+1)
            str_z = 'Zeros:'
            plt.text(0, y, str_z,
                     horizontalalignment='left',
                     verticalalignment='center')
            y = y - 0.1
            for i in range(len(zz)):
                if zz[i].imag == 0:
                    str_z = '$%+.4f$' % zz[i].real
                else:
                    str_z = (r'$%+.4f\pm \mathrm{j}%.4f$' %
                             (zz[i].real, zz[i].imag))
                plt.text(0, y, str_z,
                         horizontalalignment='left',
                         verticalalignment='center')
                y = y - 0.1

    plt.ylabel('Imag')
    plt.xlabel('Real')



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from scipy import signal
from pydsm.ir import impulse_response
from pydsm.NTFdesign.legacy import q0_from_filter_ir
from pydsm.NTFdesign.weighting import q0_weighting
from pydsm.NTFdesign import quantization_noise_gain
import scipy.linalg as la

__all__ = ["TestQ0"]


class TestQ0:

    def setUp(self):
        pass

    def test_q0_butt_bp8(self):
        # Generate filter.
        # 8th order bandpass filter
        # Freq. passed to butterworth is normalized between 0 and 1
        # where 1 is the Nyquist frequency
        fsig = 1000.
        B = 400.
        OSR = 64
        fphi = B*OSR*2
        w0 = 2*fsig/fphi
        B0 = 2*B/fphi
        w1 = (np.sqrt(B0**2+4*w0**2)-B0)/2
        w2 = (np.sqrt(B0**2+4*w0**2)+B0)/2
        hz = signal.butter(4, [w1, w2], 'bandpass', output='zpk')
        # Order
        P = 20
        # Compute q0 in two ways
        ir = impulse_response(hz, db=80)

        q0_ir = q0_from_filter_ir(P, ir)
        q0_mr = q0_weighting(P, hz)
        np.testing.assert_allclose(q0_ir, q0_mr, atol=1E-7, rtol=1E-5)

    def test_q0_butt_lp3(self):
        # Generate filter.
        # 3rd order lowpass filter
        # Freq. passed to butterworth is normalized between 0 and 1
        # where 1 is the Nyquist frequency
        B = 400.
        OSR = 256
        fphi = B*OSR*2
        w0 = 2*B/fphi
        hz = signal.butter(3, w0, 'lowpass', output='zpk')
        # Order
        P = 12
        # Compute q0 in two ways
        ir = impulse_response(hz, db=80)

        q0_ir = q0_from_filter_ir(P, ir)
        q0_mr = q0_weighting(P, hz)
        np.testing.assert_allclose(q0_ir, q0_mr, atol=1E-7, rtol=1E-5)

    def test_q0_equiv(self):
        fir = np.asarray([1.00000000e+00,  -7.63387347e-01,  -4.02004111e-01,
                          -1.53885083e-01,  -2.76434316e-04,   7.91252937e-02,
                          1.03165832e-01,   8.83276154e-02,   4.92894661e-02])
        hz = (np.asarray([-1., -1., -1.]),
              np.asarray([0.99382681+0.01056265j,  0.98780284+0.j,
                          0.99382681-0.01056265j]),
              2.2820568419526305e-07)
        gain1 = quantization_noise_gain((fir, np.hstack((1, np.zeros(8)))), hz)
        q0 = q0_weighting(8, hz)
        Q = la.toeplitz(q0)
        fir_coeff = fir.reshape((1, 9))
        gain2 = fir_coeff.dot(Q).dot(fir_coeff.T)
        np.testing.assert_allclose(gain2, gain1)



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

u"""
Synthesis of the NTF of Delta Sigma modulators (:mod:`pydsm.NTFdesign`)
=======================================================================

This module provides some strategies for the design of the Noise
Transfer Function of ΔΣ modulators. There are both functions that are
specific to PyDSM and entry points to functions in the delsig module
of PyDSM (:mod:`pydsm.delsig`).

.. currentmodule:: pydsm.NTFdesign


Key functions
-------------

.. function:: ntf_schreier()

   shorthand for :func:`delsig.ntf_schreier`

.. function:: ntf_chebyshev()

   shorthand for :func:`delsig.ntf_chebyshev`

.. function:: ntf_clans()

   shorthand for :func:`delsig.ntf_clans`

.. function:: ntf_fir_weighting()

   shorthand for :func:`weighting.ntf_fir_weighting`

.. function:: ntf_hybrid_weighting()

   shorthand for :func:`weighting.ntf_hybrid_weighting`

.. function:: ntf_fir_minmax()

   shorthand for :func:`minmax.ntf_fir_minmax`

.. function:: ntf_dunn()

   shorthand for :func:`psychoacoustic.ntf_dunn`

.. function:: ntf_fir_audio_weighting()

   shorthand for :func:`psychoacoustic.ntf_fir_audio_weighting`

.. function:: mult_weightings()

   shorthand for :func:`weighting.mult_weightings`

.. function:: quantization_noise_gain()

   shorthand for :func:`merit_factors.quantization_noise_gain`


Submodules
----------

:mod:`pydsm.NTFdesign.delsig`
  NTF synthesis functions equivalent to those in :mod:`pydsm.delsig`.

:mod:`pydsm.NTFdesign.weighting`
  NTF synthesis techniques that take as their input either a weighting
  function (indicating the cost of quantization noise power versus frequency)
  or a specification of the filter in charge of removing the quantization
  noise.

:mod:`pydsm.NTFdesign.minmax`
  NTF synthesis techniques based on a minmax approach.

:mod:`pydsm.NTFdesign.psychoacoustic`
  NTF synthesis techniques for audio modulators that result in a noise shaping
  that take into account psychoacoustics.

:mod:`pydsm.NTFdesign.merit_factors`
  Functions for determining merit factors about NTFs.

:mod:`pydsm.NTFdesign.helpers`
  Helper functions


Legacy submodule
----------------

:mod:`pydsm.NTFdesign.legacy`
  Functions that are now superseded but that are worth keeping around for
  reproducing published results


Deprecated submodules
---------------------

:mod:`pydsm.NTFdesign.filter_based`
  Alternate entry points for some functions.
"""


from .merit_factors import quantization_noise_gain
from .minmax import ntf_fir_minmax
from .delsig import ntf_schreier, ntf_chebyshev, ntf_clans
from .psychoacoustic import ntf_dunn, ntf_fir_audio_weighting
from .weighting import (ntf_fir_weighting, ntf_hybrid_weighting,
                        mult_weightings)

from .._pytesttester import PytestTester
test = PytestTester(__name__)
bench = PytestTester(__name__, mode='benchmarks')
del PytestTester



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from pydsm.ft import dtft, dtft_hermitian, idtft, idtft_hermitian

__all__ = ["TestDTFT"]


class TestDTFT:

    def setUp(self):
        pass

    def test_dtft_re10(self):
        xx = np.arange(1, 11)
        Ff = dtft(xx)
        xx2 = idtft(Ff, np.arange(10))
        np.testing.assert_allclose(xx, xx2, atol=1E-12)

    def test_dtft_re10s1(self):
        xx = np.arange(1, 11)
        Ff = dtft_hermitian(xx)
        xx2 = idtft(Ff, np.arange(10))
        np.testing.assert_allclose(xx, xx2, atol=1E-12)

    def test_dtft_re10s2(self):
        xx = np.arange(1, 11)
        Ff = dtft_hermitian(xx)
        xx2 = idtft_hermitian(Ff, np.arange(10))
        np.testing.assert_allclose(xx, xx2, atol=1E-12)



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Exceptions and warnings (:mod:`pydsm.exceptions`)
=================================================

This module defines the custom exceptions and warnings used in PyDSM

.. currentmodule:: pydsm.exceptions

Warning classes
---------------

.. autosummary::
   :toctree: generated/

    PyDsmWarning  -- base warning class
    PyDsmApproximationWarning  -- Approximation warning
    PyDsmSlowPathWarning  -- Slow path warning
    PyDsmPendingDeprecationWarning  -- Pending deprecation warning
    PyDsmDeprecationWarning  -- Deprecation warning


Error classes
-------------

No error classes are currently defined. Standard Python errors are risen.
"""


class PyDsmWarning(UserWarning):
    """Base class of warnings in PyDsm
    """
    pass


class PyDsmApproximationWarning(PyDsmWarning):
    """Class to warn about approximation in computation
    """
    pass


class PyDsmSlowPathWarning(PyDsmWarning):
    """Class to warn about slow code paths
    """
    pass


class PyDsmPendingDeprecationWarning(PyDsmWarning, PendingDeprecationWarning):
    """Class to warn about features pending deprecation
    """
    pass


class PyDsmDeprecationWarning(PyDsmWarning, DeprecationWarning):
    """Class to warn about deprecated features
    """
    pass



# -*- coding: utf-8 -*-

# Copyright (c) 2012-2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from scipy import signal
from pydsm.ir import impulse_response
from pydsm.delsig import evalTF
from pydsm.relab import db

__all__ = ["TestIR"]


class TestIR:

    def setUp(self):
        pass

    def test_butt_bp8_ir(self):
        # Generate filter.
        # 8th order bandpass filter
        # Freq. passed to butterworth is normalized between 0 and 1
        # where 1 is the Nyquist frequency
        fsig = 1000.
        B = 400.
        OSR = 64
        fphi = B*OSR*2
        w0 = 2*fsig/fphi
        B0 = 2*B/fphi
        w1 = (np.sqrt(B0**2+4*w0**2)-B0)/2
        w2 = (np.sqrt(B0**2+4*w0**2)+B0)/2
        hz = signal.butter(4, [w1, w2], 'bandpass', output='zpk')
        ir = impulse_response(hz, db=80)
        ff = np.logspace(np.log10(w0/4), np.log10(w0*4), 100)
        vv1 = evalTF(hz, np.exp(2j*np.pi*ff))
        vv2 = evalTF((ir, [1]), np.exp(2j*np.pi*ff))
        np.testing.assert_allclose(db(vv1/vv2), np.zeros_like(vv1),
                                   rtol=0, atol=3)

    def test_fir_ir(self):
        fir = np.asarray([1.0, 0.5, 0.25, 0.125])
        zpk = (np.roots(fir), np.zeros(4), 1)
        ir = impulse_response(zpk, db=80)
        np.testing.assert_allclose(fir, ir)



# -*- coding: utf-8 -*-

# Copyright (c) 2013–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the "Frequency-Domain Min-Max
# Optimization for Delta-Sigma Modulators" Matlab toolbox by Masaaki
# Nagahara (see https://www.mathworks.com/matlabcentral/fileexchange/36187)
# covered by the following copyright and permission notice
#
# Copyright (c) 2012, Masaaki Nagahara
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in
#      the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

u"""
MinMax FIR NTF design (:mod:`pydsm.NTFdesign.minmax`)
=====================================================

This modules provides a minmax FIR design method for the NTF
of ΔΣ modulators

.. currentmodule:: pydsm.NTFdesign.minmax


Functions
---------

.. autosummary::
   :toctree: generated/

   ntf_fir_minmax    -- Design FIR NTF with minmax approach


Deprecated Functions
--------------------

.. autosummary::
   :toctree: generated/

   synthesize_ntf_minmax -- Alias of ntf_fir_minmax


Notes
-----
The design strategy implemented in this module was originally proposed by
M. Nagahara and Y. Yamamoto. See the documentation of :func:`ntf_fir_minmax`
for further details.
"""

from ._fir_minmax import *     # analysis:ignore

__all__ = _fir_minmax.__all__ + ['test']

from ..._pytesttester import PytestTester
test = PytestTester(__name__)
del PytestTester



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..sets import *
from ..utils import *
from ..interface import *
from ..arrays import cvxpy_matrix

# Geometric Mean
def geo_mean(x):
    r"""
    | :math:`\mbox{geo\_mean} :
      \mathbb{R}_+^n \\to \mathbb{R},
      \ \mbox{geo\_mean}(x) = \\big( \prod_{i=1}^n x_i \\big)^{1/n}`.
    | Concave and increasing.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: number or
             :ref:`tree<tree_obj>`.
    """

    # Process input
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        x = vstack([x])
    elif (type(x) is not cvxpy_matrix and
          type(x).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid argument type')

    # x must be a vector
    (m,n) = x.shape
    if n!=1:
        raise ValueError('Invalid argument dimension')

    # Check nonnegativity if cvxpy_matrix
    if (type(x) is cvxpy_matrix and
        not np.all(x >= 0)):
        return -np.inf

    # Construct program
    z = variable(m,1)
    v = variable(m,1)
    t = variable()
    p = program(maximize(t),
                [belongs(vstack((v,t)),geo_mean_cone),
                 greater_equals(z,v), geq(v,0)],
                [z],
                name='geo_mean')
    return p(x)



# Copyright © 2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.


# Based on code Copyright (c) 2005-2024, NumPy Developers.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

#    * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.

#    * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.

#    * Neither the name of the NumPy Developers nor the names of any
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Pytest test runner.

This module implements the ``test()`` function for PyDSM modules and is
based on the NumPy module with the same name

Warnings filtering and other runtime settings should be dealt with in the
``pytest.ini`` file in the repo root. The behavior of the test depends on
whether or not that file is found as follows:

* ``pytest.ini`` is present (develop mode)
    All warnings except those explicitly filtered out are raised as error.
* ``pytest.ini`` is absent (release mode)
    DeprecationWarnings and PendingDeprecationWarnings are ignored, other
    warnings are passed through.

In practice, tests run from the repo are run in development mode from an
inplace build with ``pytest pydsm``.

This module is imported by every subpackage.
"""
import sys
import os

__all__ = ['PytestTester']


def _show_pydsm_info():
    import pydsm
    print(f"PyDSM version {pydsm.__version__}")

class PytestTester:
    """
    Pytest test runner.

    Calling this test function finds and runs all tests associated with the
    module and all its sub-modules.

    Attributes
    ----------
    module_name : str
        The name of the (sub)package/module to test.

    Parameters
    ----------
    module_name : module name
        The name of the (sub)package/module to test.

    Notes
    -----
    This class is not publicly exposed.

    """

    def __init__(self, module_name, mode='tests'):
        self.module_name = module_name
        assert mode in ['tests', 'benchmarks']
        self.mode = mode

    def __call__(self, label='fast', verbose=1, extra_argv=None,
                 doctests=False, coverage=False, durations=-1, tests=None):
        """
        Run tests for module using pytest.

        Parameters
        ----------
        label : {'fast', 'full'}, optional
            Identifies the tests to run. When set to 'fast', tests decorated
            with `pytest.mark.slow` are skipped, when 'full', the slow marker
            is ignored.
        verbose : int, optional
            Verbosity value for test outputs, in the range 1-3. Default is 1.
        extra_argv : list, optional
            List with any extra arguments to pass to pytests.
        doctests : bool, optional
            Perform doctests
        coverage : bool, optional
            If True, report coverage. Default is False.
            Requires installation of (pip) pytest-cov.
        durations : int, optional
            If < 0, do nothing, If 0, report time of all tests, if > 0,
            report the time of the slowest `timer` tests. Default is -1.
        tests : test or list of tests
            Tests to be executed with pytest '--pyargs'

        Returns
        -------
        result : bool
            Return True on success, false otherwise.

        Notes
        -----
        Each PyDSM module exposes `test` in its namespace to run all tests for
        it. E.g. `pydsm.NTFdesign.test()`.
        """

        import pytest
        import warnings

        module = sys.modules[self.module_name]
        module_path = os.path.abspath(module.__path__[0])

        # setup the pytest arguments
        pytest_args = ["-l"]

        # offset verbosity. The "-q" cancels a "-v".
        pytest_args += ["-q"]

        if self.mode == 'tests':
            pytest_args += ["--ignore-glob", "**/benchmark_*.py"]
        else:
            pytest_args += ["--ignore-glob", "**/test_*.py"]

        if doctests:
            pytest_args += ["--doctest-modules"]

        if extra_argv:
            pytest_args += list(extra_argv)

        if verbose > 1:
            pytest_args += ["-" + "v"*(verbose - 1)]

        if coverage:
            pytest_args += ["--cov=" + module_path]

        if label == "fast":
            pytest_args += ["-m", "not slow"]

        elif label != "full":
            pytest_args += ["-m", label]

        if durations >= 0:
            pytest_args += [f"--durations={durations}"]

        if tests is None:
            tests = [self.module_name]

        pytest_args += ["--pyargs"] + list(tests)

        # run tests.
        _show_pydsm_info()

        try:
            code = pytest.main(pytest_args)
        except SystemExit as exc:
            code = exc.code

        return code == 0



# -*- coding: utf-8 -*-

# Copyright (c) 2015, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import division

import numpy as np

__all__ = ["axisLabels"]


def axisLabels(points, incr):
    """
    Generate alphanumeric axis labels.

    If the parameter `incr` is an integer, the function returns an array
    of strings corresponding to the values specified in `points`,
    stepping through them with indexes 0, incr, 2*incr, ...

    If the parameter `incr` is a couple of integers `(start, incr)`,
    the stepping occurs with indexes start, start+incr, start+2*incr, ...

    Parameters
    ----------
    points : array like of floats
        the axis points to label
    incr : int or couple of ints
        the increment to use in labelling

    Returns
    -------
    s : list of strings
        list of labels

    Notes
    -----

    In the original DELSIG function, the parameter `points` is named `range`.

    Values in `points` less than 1e-6 are truncated to zero. This value is
    hardwired in the function.
    """
    points = np.asarray(points).flatten()
    # Flatten assures that a copy is taken, so that the following assignement
    # does not alter the original vector
    points[np.abs(points) < 1e-6] = 0
    if isinstance(incr, int):
        start = 0
        step = incr
    elif (hasattr(incr, '__len__') and
          len(incr) == 2 and
          np.all([isinstance(x, int) for x in incr])):
        start = incr[0]
        step = incr[1]
    else:
        raise TypeError('Parameter incr must be integer or couple of integer '
                        'values')
    return ['%g' % points[i] for i in range(start, len(points), step)]



# -*- coding: utf-8 -*-

# Copyright (c) 2013, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from warnings import warn
from ...exceptions import PyDsmDeprecationWarning
from ...utilities import digested_options

__all__ = ['ntf_fir_minmax', 'synthesize_ntf_minmax']


def ntf_fir_minmax(order=32, osr=32, H_inf=1.5, f0=0, zf=False,
                   **options):
    u"""Synthesize FIR NTF for LP or BP ΔΣ modulator by min-max optimization.

    The strategy aims at minimizing the peak value of the NTF in the signal
    band, while respecting the Lee criterion.

    Parameters
    ----------
    order : int, optional
        Order of the delta-sigma modulator. Defaults to 32.
    osr : real or array of reals, optional
        The oversampling ratio, based on the input signal bandwidth
        Defaults to 32. If there are multiple signal bands, multiple
        OSRs can be provided, one for each of them.
    H_inf : real, optional
        Max allowed peak value of the NTF. Used to enforce the Lee criterion.
        Defaults to 1.5.
    f0 : real, or array of reals, optional
        The normalized center frequency of the modulator. Value in [0,1/2].
        1/2 indicates half the sample frequency. Defaults to 0, indicating an
        LP modulator. An array of values can be provided for multiband design.
    zf : bool, optional
        Flag controlling the pre-assignement of NTF zeros. If ``False``, the
        design is practiced without any zero pre-assignment. If ``True``, a
        zero is pre-assigned at the modulator center-band. Defaults to False.

    Returns
    -------
    ntf : tuple
        noise transfer function in zpk form.

    cvxpy_opts : dictionary, optional
       A dictionary of options to use with the ``cvxpy`` modeling library.
       Allowed options include:

       ``override_kktsolver`` (bool)
           Whether to override the default ``cvxopt`` kkt solver using the
           ``chol`` kkt solver.
           Leave this at the default False setting, to avoid errors.
       ``solver`` (string)
           The solver backend to use. Either `cvxopt` or `scs`

    cvxopt_opts : dict, optional
        A dictionary of options for the ``cvxopt`` optimizer.
        Allowed options include:

        ``maxiters`` (int)
            Maximum number of iterations
        ``abstol`` (real)
            Absolute accuracy
        ``reltol`` (real)
            Relative accuracy
        ``feastol`` (real)
            Tolerance for feasibility conditions (defaults to 1e-6)

        Do not use other options since they could break ``cvxopt`` in
        unexpected ways. These options can be passed when using the
        ``cvxpy_old`` modeler, the ``picos`` modeler or the ``cvxpy`` modeler
        with the ``cvxopt`` backend.
    scs_opts : dict, optional
        A dictionary of options for the ``scs`` optimizer.  Allowed options
        include:

        ``max_iters`` (int)
            Maximum number of iterations
        ``eps`` (real)
            Convergence tolerance
        ``alpha`` (real)
            Relaxation parameter
        ``normalize`` (bool)
            Whether to precondition data matrices
        ``use_indirect`` (bool)
            Whether to use indirect solver for KKT sytem (instead of direct)

       Do not use other options since they could break ``scs`` in
       unexpected ways. These options can be passed when using the
       ``cvxpy`` modeler with the ``scs`` backend.

    Notes
    -----
    The design strategy implemented in this module is described in the paper
    [1]_.

    Default values for the options not directly documented in the function
    call signature can be checked and updated by changing the function
    ``default_options`` attribute.

    .. [1] M. Nagahara and Y. Yamamoto, *Frequency-Domain Min-Max Optimization
       of Noise-Shaping Delta-Sigma Modulators*, IEEE Trans. SP, vol. 60 n. 6
       June 2012.

    For more information on the ``CVXOPT`` optimizer parameters, see
    the corresponding documentation. Similarly, for more infomration
    on the ``scs`` optimizer see its documentation. Finally, for more
    iniformation on the parameters controlling the optimization
    modeler and frontend ``cvxpy``, see its documentation.

    """
    # Manage optional parameters
    opts = digested_options(
        options, ntf_fir_minmax.default_options,
        ['show_progress', 'modeler'], [], False)
    dig_opts = {'show_progress': opts['show_progress'],
                'cvxpy_opts': {},
                'cvxpy_tdr_opts': {},
                'picos_opts': {}}
    if opts['modeler'] == 'cvxpy':
        opts.update(digested_options(
            options, ntf_fir_minmax.default_options,
            [], ['cvxpy_opts'], False))
        if opts['cvxpy_opts']['solver'] == 'cvxopt':
            dig_opts['cvxpy_opts'].update(digested_options(
                options, ntf_fir_minmax.default_options,
                [], ['cvxopt_opts'], False)['cvxopt_opts'])
            if opts['cvxpy_opts'].get('override_kktsolver', True):
                dig_opts['cvxpy_opts']['kktsolver'] = 'chol'
        elif opts['cvxpy_opts']['solver'] == 'scs':
            dig_opts['cvxpy_opts'].update(digested_options(
                options, ntf_fir_minmax.default_options,
                [], ['scs_opts'], False)['scs_opts'])
        opts['cvxpy_opts'].pop('override_kktsolver')
        dig_opts['cvxpy_opts'].update(opts['cvxpy_opts'])
        from ._fir_minmax_cvxpy import (
            ntf_fir_from_digested as _ntf_fir_from_digested)
    elif opts['modeler'] == 'cvxpy_old':
        dig_opts['cvxpy_tdr_opts'].update(digested_options(
            options, ntf_fir_minmax.default_options,
            [], ['cvxopt_opts'], False)['cvxopt_opts'])
        from ._fir_minmax_cvxpy_tdr import (
            ntf_fir_from_digested as _ntf_fir_from_digested)
    elif opts['modeler'] == 'picos':
        dig_opts['picos_opts'].update(digested_options(
            options, ntf_fir_minmax.default_options,
            [], ['cvxopt_opts'], False)['cvxopt_opts'])
        from ._fir_minmax_picos import (
            ntf_fir_from_digested as _ntf_fir_from_digested)
    else:
        raise ValueError('Unsupported modeling backend {}'.format(
            opts['modeler']))
    digested_options(options, {})
    if np.isscalar(f0):
        f0 = [f0]
    if np.isscalar(osr):
        osr = len(f0) * [osr]
    if not len(osr) == len(f0):
        raise ValueError('Incorrect multiband specification')
    # Do the computation
    ntf_ir = _ntf_fir_from_digested(order, osr, H_inf, f0, zf, **dig_opts)
    return (np.roots(ntf_ir), np.zeros(order), 1.)

ntf_fir_minmax.default_options = {"cvxpy_opts": {'override_kktsolver': False,
                                                 'solver': 'cvxopt'},
                                  "cvxopt_opts": {'maxiters': 100,
                                                  'abstol': 1e-7,
                                                  'reltol': 1e-6,
                                                  'feastol': 1e-6},
                                  'scs_opts': {'max_iters': 25000,
                                               'eps': 1e-10,
                                               'alpha': 1.8,
                                               'normalize': True,
                                               'use_indirect': False},
                                  'show_progress': True,
                                  'modeler': 'cvxpy_old'}


# Following part is deprecated


def synthesize_ntf_minmax(order=32, osr=32, H_inf=1.5, f0=0, zf=False,
                          **options):
    """
    Alias of :func:`ntf_fir_minmax`

    .. deprecated:: 0.11.0
       Function is now available from the :mod:`pydsm.NTFdesign`
       module with name :func:`ntf_fir_minmax`

    """
    warn("Function superseded by ntf_fir_minmax in "
         "NTFdesign module", PyDsmDeprecationWarning)
    return ntf_fir_minmax(order, osr, H_inf, f0, zf, **options)



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Computation of the impulse response of a DT filter (:mod:`pydsm.ir`)
====================================================================

Compute (approximating by truncation) the impulse response
of a discrete time filter, trying to (roughly) guess the appropriate
truncation.

.. currentmodule:: pydsm.ir

Functions
---------

.. autosummary::
   :toctree: generated/

    guess_ir_length  -- Guess appropriate truncation length
    impulse_response  -- Compute impulse response of DT filter
"""

from __future__ import division, print_function

import numpy as np
import scipy as sp
__import__("scipy.signal")

__all__ = ["impulse_response", "guess_ir_length"]


def impulse_response(h, m=None, db=80):
    """
    Computes the filter impulse response

    Parameters
    ----------
    h : tuple_like
        the filter definition either in zpk or in nd form.

    Returns
    -------
    ir : ndarray
        the truncated impulse response

    Other Parameters
    ----------------
    m : int, optional
        the number of samples after which the impulse response should be
        truncated. Defaults to None, which means *try to guess*
    db : real, optional
        a hint about how to guess the length where the impuls response
        should be truncated (defaults to 80)

    Notes
    -----
    The guess about the lenght where the impulse response can be truncated
    is extremely rough. See :func:`guess_ir_length` in this module for
    further info.
    """
    if len(h) == 3:
        (b, a) = sp.signal.zpk2tf(*h)
        b = b.real
        a = a.real
    else:
        (b, a) = h
    if m is None:
        m = guess_ir_length(h, db)
    ins = np.zeros(m)
    ins[0] = 1
    return sp.signal.lfilter(b, a, ins)


def guess_ir_length(h, db=80):
    """
    Tries to estimate an appropriate length for the filter response

    Parameters
    ----------
    h : tuple_like
        the filter definition either in zpk or in nd form.
    db : real, optional
        a hint about how to guess the length where the impulse response
        should be truncated. This is defined on a log scale. The larger
        the longer the resulting length. Defaults to 80.

    Returns
    -------
    m : int
        a guess about the appropriate number of samples to represent the
        filter impulse response with the required accuracy

    Notes
    -----
    The guess is based on the slowlest pole of the filter, considering when
    its response is attenuated to -db. This can be by far too optimistic in
    some cases and particularly when there are overlapping or similar poles.

    Do not try to use this function for filters with poles in 1.
    """
    # Put h in zpk form if it is in tf form
    if len(h) == 2:
        h = sp.signal.tf2zpk(*h)
    pp = h[1]
    t_z = len(h[0])+1
    if len(pp) == 0:
        t_p = 0
    else:
        # Try to estimate length of decay of the filter h.
        # The estimation is extremely rough, based on the decay
        # rate of the pole with maximum magnitude.
        # Thus, it breaks easily when there are poles very close
        # one to the other or overlapping.
        # Furthermore, this code should not be called if there is
        # a pole in z=1.
        os = np.seterr(divide='ignore')
        sr = np.log(np.abs(pp))
        np.seterr(**os)
        # Take slowlest pole
        wmin = np.min(np.abs(sr))
        # 1/omega min is time constant in sample periods.
        # Let's multiply the time constant in order to have
        # the transient attenuated by db decibels
        t_p = int(np.ceil(db/20*np.log(10)/wmin))
    return t_p+t_z



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Simulation of Delta Sigma modulators (:mod:`pydsm.simulation`)
==============================================================

This module provides some functions for the simulation of of ΔΣ
modulators. Currently, the functions included in this module are just
alternative entry points for functions in the :mod:`pydsm.delsig` module.

.. currentmodule:: pydsm.simulation


Functions
---------

.. autosummary::
   :toctree: generated/

   simulateDSM   -- Delta sigma modulator simulation
   ds_quantize   -- quantization function
"""

# Promote some functions/global variables to the simulation namespace
from .delsig import simulateDSM
from .delsig import ds_quantize

__all__ = ['simulateDSM', 'ds_quantize']



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np

from pydsm.NTFdesign.helpers import spread_fir_uc_zeros
from pydsm.NTFdesign.merit_factors import quantization_noise_gain
from pydsm.relab import cplxpair
from pydsm.delsig import ds_optzeros


class TestNTFdesignHelpers:
    def setUp(self):
        pass

    def test_uc_zeros(self):
        order = 7
        OSR = 32
        zeros1 = spread_fir_uc_zeros(order, OSR,
                                     quantization_noise_gain,
                                     cf_kwargs={'bounds': (0, 0.5/OSR)})
        zeros1 = cplxpair(zeros1)
        zeros2 = ds_optzeros(order)
        zeros2 = np.exp(1j*np.pi*zeros2/OSR)
        zeros2 = cplxpair(zeros2)
        np.testing.assert_almost_equal(zeros1, zeros2, 4)



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

from matplotlib import pyplot as plt
from pydsm.delsig import synthesizeNTF
from pydsm.delsig import plotPZ

import logging
logging.getLogger('matplotlib').setLevel(logging.ERROR)

__all__ = ["TestPlotPZ"]


class TestPlotPZ:

    def setUp(self):
        self.saved_backend = plt.get_backend()
        plt.switch_backend('Agg')

    def tearDown(self):
        plt.switch_backend(self.saved_backend)

    def test_default(self):
        ntf = synthesizeNTF()
        plotPZ(ntf)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from .call_solver import call_solver

# Function
def solve_prog(p,quiet):
    """
    Solves optimization program.

    :param p: cvxpy_program
    """

    # Expand
    p_expanded = p._get_expanded_program()

    # Compute signs
    if p.action == MINIMIZE:
        sign = 1.
    else:
        sign = -1.

    # Solve convex relaxation
    if not quiet:
        print('\nCalling CVXOPT ...')
    sol = call_solver(p_expanded,quiet)

    valid = True

    # Primal infeasible
    if sol['status'] == PRIMAL_INFEASIBLE:
        obj = np.inf*sign

    # Uknown
    elif sol['status'] == UNKNOWN:
        obj = sol[PRIMAL_OBJECTIVE]*sign
        valid = False

    # Dual infeasible
    elif sol['status'] == DUAL_INFEASIBLE:
        obj = np.inf*(-sign)

    # Optimal
    else:
        obj =sol[PRIMAL_OBJECTIVE]*sign

    # Return optimal value
    return obj,valid



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""PyDSM Toolbox for the design and simulation of ΔΣ modulators
============================================================

PyDSM © 2012–2025 Sergio Callegari. All rights reserved.

Some code portions confined in the :mod:`pydsm.delsig` module are
ported or reimplemented from the DELSIG_ toolbox for MATLAB,
copyrighted by R. Schreier.

PyDSM focuses on the design of Noise Transfer Functions for Digital ΔΣ
Modulators and in the simulation of the resulting modulators.


Main modules
------------

.. toctree::
   :maxdepth: 1

   pydsm.delsig
   pydsm.NTFdesign
   pydsm.simulation
   pydsm.audio_weightings
   pydsm.iso226

Utility modules
---------------

.. toctree::
   :maxdepth: 1

   pydsm.correlations
   pydsm.ft
   pydsm.ir
   pydsm.relab
   pydsm.utilities


Error handling
--------------

.. toctree::
   :maxdepth: 1

   pydsm.exceptions

.. _DELSIG:
   https://www.mathworks.com/matlabcentral/fileexchange/19-delta-sigma-toolbox

"""

# Read version info
from ._version import __version__

# Promote some key functions to the pydsm namespace
from . import delsig
from . import NTFdesign
from . import simulation
from . import audio_weightings
from . import iso226

from ._pytesttester import PytestTester
test = PytestTester(__name__)
bench = PytestTester(__name__, mode='benchmarks')
del PytestTester



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

u"""
Output filter based NTF synthesis (:mod:`pydsm.NTFdesign.filter_based`)
=======================================================================

This modules provides code for the synthesis of the modulator NTF, based
on the filter placed after the modulator for the removal of the quantization
noise.

.. deprecated:: 0.11.0
    Use `ntf_fir_weighting` or functions from `NTFdesign.weighting`
    module.

.. currentmodule:: pydsm.NTFdesign.filter_based


Deprecated functions
--------------------

.. autosummary::
   :toctree: generated/

   quantization_noise_gain -- Alternate quantization noise power gain func
   quantization_noise_gain_by_conv -- Alternate function based on convolution
   q0_from_filter -- Alternate compute Q matrix from the output filter
   synthesize_ntf_from_filter -- Alternate compute FIR NTF from output filter
"""

from __future__ import division, print_function

from .merit_factors import quantization_noise_gain as _quantization_noise_gain
from .legacy import (quantization_noise_gain_by_conv as
                     _quantization_noise_gain_by_conv,
                     q0_from_filter_ir)
from .weighting import q0_weighting, ntf_fir_from_q0
from warnings import warn
from ..exceptions import PyDsmDeprecationWarning
from ..utilities import digested_options

__all__ = ["quantization_noise_gain", "quantization_noise_gain_by_conv",
           "q0_from_filter", "synthesize_ntf_from_filter"]


def quantization_noise_gain(NTF, H, H_type='zpk', **options):
    r"""Compute the quantization noise power gain after a filter

    .. deprecated:: 0.11.0
       Use :func:`pydsm.NTFdesign.quantization_noise_gain`. Note that the
       interface is slightly different.

    Parameters
    ----------
    NTF : tuple
        NTF definition in zpk or nd form
    H : tuple or callable or array_like
        output filter definition in zpk or ba form if H_type='zpk' or 'ba'
        (in this case, H is a tuple with 3 or 2 entries);
        output filter magnitude response if H_type='mag' (in this case, H is
        a callable with argument f in [0,1/2]);
        output filter impulse response if H_type='imp' (in this case, H is an
        array)
    H_type : str
        type of specification for parameter H. One of: 'zpk', 'ba', 'mag' or
        'imp'

    Returns
    -------
    a : real
        noise power gain

    Other parameters
    ----------------
    quad_opts : dictionary, optional
        Parameters to be passed to the ``quad`` function used internally as
        an integrator. Allowed options are ``epsabs``, ``epsrel``, ``limit``,
        ``points``. Do not use other options since they could break the
        integrator in unexpected ways. Defaults can be set by changing the
        function ``default_options`` attribute.

    Raises
    ------
    ValueError
        'Incorrect filter type specification' if the filter is not specified
        correctly.

    Notes
    -----
    In the default case the computation is practiced as

    .. math::
        2\int_{f=0}^{\frac{1}{2}}
        \left|H\left(\mathrm{e}^{\mathrm{i} 2\pi f}\right)\right|^2
        \left|\mathit{NTF}
        \left(\mathrm{e}^{\mathrm{i} 2\pi f}\right)\right|^2 df

    See Also
    --------
    scipy.integrate.quad : for the meaning of the integrator parameters.
    """
    warn("Function meant for removal", PyDsmDeprecationWarning)
    # Manage optional parameters
    opts = digested_options(options, quantization_noise_gain.default_options,
                            [], ['quad_opts'])
    if H_type == 'zpk' or H_type == 'ba':
        w = H
    elif H_type == 'imp':
        w = (H, [1])
    elif H_type == 'mag':
        w = lambda f: H(f)**2
    else:
        raise ValueError("Incorrect filter type specification")
    return _quantization_noise_gain(NTF, w, **opts)

quantization_noise_gain.default_options = \
    _quantization_noise_gain.default_options.copy()


def quantization_noise_gain_by_conv(NTF, H, H_type='zpk', db=80):
    """
    Alias of :func:`pydsm.NTFdesign.legacy.quantization_noise_gain_by_conv`

    .. deprecated:: 0.11.0
        Function has been moved to the :mod"`NTFdesign.legacy` module.
    """
    warn("Function moved in ``NTFdesign.legacy`` module",
         PyDsmDeprecationWarning)
    return _quantization_noise_gain_by_conv(NTF, H, H_type, db)


def q0_from_filter(P, H, H_type='zpk', **options):
    """
    Compute Q matrix from the modulator output filter

    .. deprecated:: 0.11.0
       Use :func:`pydsm.NTFdesign.weighting.q0_weighting`. Note that the
       interface is slightly different.

    Parameters
    ----------
    P : int
        order of the FIR to be eventually synthesized
    H : tuple or array_like or callable
        output filter description.
        This is given by a zpk or ba form if F_type is 'zpk' or 'ba'.
        It is a magnitude response (function of f, with f in [0,1/2]) if
        F_type is 'mag'.
        It is an impulse response if F_type is 'imp'.
    H_type : str
        string indicating how F is expressed. Can be 'zpk', 'ba', 'mag' or
        'imp'

    Returns
    -------
    q0 : ndarray
        the first row of the matrix Q used in the NTF optimization

    Other parameters
    ----------------
    quad_opts : dictionary, optional
        Parameters to be passed to the ``quad`` function used internally as
        an integrator. Allowed options are ``epsabs``, ``epsrel``, ``limit``,
        ``points``. Do not use other options since they could break the
        integrator in unexpected ways. Defaults can be set by changing the
        function ``default_options`` attribute.

    Raises
    ------
    ValueError
        'Incorrect filter type specification' if the filter is not specified
        correctly.

    Notes
    -----
    The Q matrix being synthesized has (P+1) times (P+1) entries.

    See Also
    --------
    scipy.integrate.quad : for the meaning of the integrator parameters.
    """
    warn("Function meant for removal", PyDsmDeprecationWarning)
    # Manage optional parameters
    opts = digested_options(options, q0_from_filter.default_options,
                            [], ['quad_opts'])
    # Do the computation
    if H_type == 'zpk' or H_type == 'ba':
        w = H
        q0 = q0_weighting(P, w, **opts)
    elif H_type == 'imp':
        q0 = q0_from_filter_ir(P, H)
    elif H_type == 'mag':
        w = lambda f: H(f)**2
        q0 = q0_weighting(P, w, **opts)
    else:
        raise ValueError("Incorrect filter type specification")
    return q0

q0_from_filter.default_options = q0_weighting.default_options.copy()


def synthesize_ntf_from_filter(order, H, H_type='zpk', H_inf=1.5,
                               normalize="auto", **options):
    u"""Synthesize a FIR NTF based on the ΔΣM output filter.

    The ΔΣ modulator NTF is designed after a specification of the
    filter in charge of removing the quantization noise

    .. deprecated:: 0.11.0
       Use :func:`pydsm.NTFdesign.ntf_fir_weighting`. Note that the interface
       is slightly different.

    Parameters
    ----------
    order : int
        Delta sigma modulator order
    F : tuple or callable or array_like
        filter specification, the format depends on parameter F_type.
        a zpk or ba tuple if F_type is 'zpk' or 'ba', respectively.
        a function of f, for f in [0,1/2] if F_type is 'mag'
        an array containing an impulse response if F_type is 'imp'
    F_type : str
        string indicating the type of filter specification. Can be 'zpk',
        'ba', 'mag' or 'imp'.
    H_inf : real, optional
        Max peak NTF gain, defaults to 1.5, used to enforce the Lee criterion
    normalize : string or real, optional
        Normalization to apply to the quadratic form used in the NTF
        selection. Defaults to 'auto' which means setting the top left entry
        in the matrix Q defining the quadratic form to 1.

    Returns
    -------
    ntf : ndarray
        FIR NTF in zpk form

    Other parameters
    ----------------
    show_progress : bool, optional
        provide extended output, default is True and can be updated by
        changing the function ``default_options`` attribute.
    cvxopt_opts : dictionary, optional
        A dictionary of options for the ``cvxopt`` optimizer
        Allowed options include:

        ``maxiters``
            Maximum number of iterations (defaults to 100)
        ``abstol``
            Absolute accuracy (defaults to 1e-7)
        ``reltol``
            Relative accuracy (defaults to 1e-6)
        ``feastol``
            Tolerance for feasibility conditions (defaults to 1e-6)

        Do not use other options since they could break `cvxopt` in
        unexpected ways. Defaults can be set by changing the function
        ``default_options`` attribute.
    quad_opts : dictionary, optional
        Parameters to be passed to the ``quad`` function used internally as
        an integrator. Allowed options are ``epsabs``, ``epsrel``, ``limit``,
        ``points``. Do not use other options since they could break the
        integrator in unexpected ways. Defaults can be set by changing the
        function ``default_options`` attribute.

    See Also
    --------
    scipy.integrate.quad : for the meaning of the integrator parameters

    Notes
    -----
    Cosider the `CVXOPT User’s Guide <https://cvxopt.org/userguide/>`_
    for the optimizer parameters.
    """
    warn("Function meant for removal", PyDsmDeprecationWarning)
    # Manage optional parameters
    opts1 = digested_options(options,
                             synthesize_ntf_from_filter.default_options,
                             ['quad_opts'], [], False)
    opts2 = digested_options(options,
                             synthesize_ntf_from_filter.default_options,
                             ['show_progress'], ['cvxopt_opts'])
    # Do the computation
    q0 = q0_from_filter(order, H, H_type, **opts1)
    return ntf_fir_from_q0(q0, H_inf, normalize, **opts2)

synthesize_ntf_from_filter.default_options = \
    q0_from_filter.default_options.copy()
synthesize_ntf_from_filter.default_options.update(
    ntf_fir_from_q0.default_options)



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Transfer function management a la DELSIG.
=========================================
"""

import numpy as np

__all__ = ["evalTF", "evalRPoly"]


def evalTF(tf, x):
    """
    Evaluates a transfer function.

    This function can be used either for discrete time or continuous time
    transfer functions (see the notes).

    Parameters
    ----------
    tf : tuple
        transfer function in zpk or ba form
    x : complex or array_like of complex
        value or vector of values where the tf is to be evaluated

    Returns
    -------
    y : ndarray
        value(s) of transfer function at the given complex values.

    Notes
    -----
    Parameter x corresponds to 's' or 'z' in CT or DT transfer functions
    respectively. Thus it should be 1j*omega or exp(1j*omega*T).

    With respect to the tf parameter, zpk form is a triple containing a list
    of zeros, a list of poles and a scalar gain. The ba form (also called tf
    form in the scipy documentation) is a couple containing a list of the
    numerator coefficients and a list of the denominator coefficients. The
    coefficients are sorted from the higher power of 's' or 'z' to the lower,
    so that the last coefficient is in fact the constant term in the
    numerator/denominator polynomial.
    """
    if len(tf) == 3:
        retval = np.asarray(evalRPoly(tf[0], x, tf[2]))/evalRPoly(tf[1], x, 1)
    elif len(tf) == 2:
        retval = np.asarray(np.polyval(tf[0], x))/np.polyval(tf[1], x)
    if retval.size == 1:
        return retval.item()
    return retval


def evalRPoly(roots, x, k=1):
    """
    Compute the value of a polynomial that is given in terms of its roots.
    Roots at infinity are removed before the computation.

    Parameters
    ----------
    roots : array_like
        roots of polynomial
    x : complex or array_like of complex
        complex value or vector of complex values where tf is to be
        evaluated
    k -> real, optional
        gain, defaults to 1.

    Returns
    -------
    y : ndarray
        value of polynomial at the given complex values.
    """
    y = k*np.ones_like(x)
    roots = np.asarray(roots)
    # Remove roots at infinity
    roots = roots[np.logical_not(np.isinf(roots))]
    for i in range(roots.size):
        y = y*(x-roots[i])
    return y



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from .defs import *
from functools import reduce

#***********************************************************************#
# Class definition: cvxpy_obj                                           #
#***********************************************************************#
class cvxpy_obj(object):

    # Method: __init__
    def __init__(self,object_type,value,name):
        """
        Class constructor.

        :param object_type: Keyword (See cvxpy.def).
        :param value: Number.
        :param name: String.
        """

        self.type = object_type
        self.value = value
        self.name = name
        self.shape = (1,1)
        self.variables = cvxpy_list()
        self.parameters = cvxpy_list()
        self.T = self

    def __lt__(self, other):
        return self.name < other.name

    # Method: is_convex
    def is_convex(self):
        """
        Determines if object follows DCP rules of convexity.
        """

        return True

    # Method: is_concave
    def is_concave(self):
        """
        Determines if object follows DCP rules of concavity.
        """

        return True

    # Method: is_dcp
    def is_dcp(self):
        """
        Determines if object is DCP-compliant.
        """

        return True

    # Method: is_affine
    def is_affine(self):
        """
        Determines if object is affine.
        """

        return True

    # Method: is_nonnegative_constant
    def is_nonnegative_constant(self):
        """
        Determines if object is a nonnegative constant.
        """

        return self.value >= 0

    # Method: is_nonpositive_constant
    def is_nonpositive_constant(self):
        """
        Determines if object is a nonpositive constant.
        """

        return self.value <= 0

    # Method: __add__
    def __add__(self,other):

        return self.__addsub__(other,SUMMATION)

    # Method: __sub__
    def __sub__(self,other):

        return self.__addsub__(other,SUBTRACTION)

    # Method: __radd__
    def __radd__(self,other):
        return self.__raddsub__(other,SUMMATION)

    # Method: __rsub__
    def __rsub__(self,other):

        return self.__raddsub__(other,SUBTRACTION)

    # Method: __addsub__
    def __addsub__(self,other,op):
        """
        Handles left add and left subtract.

        :param other: Right hand side.
        :param op: Keyword (See cvxpy.defs)
        """

        # Create operator
        operator = cvxpy_obj(OPERATOR,np.nan,SUMMATION)

        # Create args to combine
        if (type(self) is cvxpy_tree and
            self.item.name == SUMMATION):
            args = self.children
        else:
            args = [self]

        # Number
        if np.isscalar(other):
            if other == 0.0:
                return self
            if op == SUMMATION:
                constant = cvxpy_obj(CONSTANT,other,str(other))
            else:
                constant = cvxpy_obj(CONSTANT,-other,str(-other))
            return cvxpy_tree(operator,args+[constant])

        # Scalar variable or scalar param
        elif (type(other) is cvxpy_scalar_var or
              type(other) is cvxpy_scalar_param):
            if op == SUMMATION:
                return cvxpy_tree(operator,args+[other])
            else:
                return cvxpy_tree(operator,args+[-other])

        # Tree
        elif type(other) is cvxpy_tree:
            if other.item.name == SUMMATION:
                if op == SUMMATION:
                    return cvxpy_tree(operator,
                                      args+other.children)
                else:
                    return cvxpy_tree(operator,
                                      args+[-x for x in other.children])
            else:
                if op == SUMMATION:
                    return cvxpy_tree(operator,args+[other])
                else:
                    return cvxpy_tree(operator,args+[-other])

        # Matrix
        elif type(other) is cvxpy_matrix:
            (m,n) = other.shape
            new_ar = cvxpy_array(m,n)
            for i in range(0,m,1):
                for j in range(0,n,1):
                    if op == SUMMATION:
                        new_ar[i,j] = self+other[i,j]
                    else:
                        new_ar[i,j] = self-other[i,j]
            return new_ar

        # Not implemented
        else:
            return NotImplemented

    # Method: __raddsub__
    def __raddsub__(self,other,op):
        """
        Handles right add and right subtract.

        :param other: Left hand side.
        :param op: Keyword (See cvxpy.defs)
        """

        # Create operator
        operator = cvxpy_obj(OPERATOR,np.nan,SUMMATION)

        # Create args to combine
        if (type(self) is cvxpy_tree and
            self.item.name == SUMMATION):
            if op == SUMMATION:
                args = self.children
            else:
                args = [-x for x in self.children]
        else:
            if op == SUMMATION:
                args = [self]
            else:
                args = [-self]

        # Number
        if np.isscalar(other):
            if other == 0.0:
                if op == SUMMATION:
                    return self
                else:
                    return -self
            constant = cvxpy_obj(CONSTANT,other,str(other))
            return cvxpy_tree(operator,[constant]+args)

        # Scalar variable or scalar param
        elif (type(other) is cvxpy_scalar_var or
              type(other) is cvxpy_scalar_param):
            return cvxpy_tree(operator,[other]+args)

        # Tree
        elif type(other) is cvxpy_tree:
            if other.item.name == SUMMATION:
                return cvxpy_tree(operator,other.children+args)
            else:
                return cvxpy_tree(operator,[other]+args)

        # Matrix
        elif type(other) is cvxpy_matrix:
            (m,n) = other.shape
            new_ar = cvxpy_array(m,n)
            for i in range(0,m,1):
                for j in range(0,n,1):
                    if op == SUMMATION:
                        new_ar[i,j] = other[i,j]+self
                    else:
                        new_ar[i,j] = other[i,j]-self
            return new_ar

        # Not implemented
        else:
            return NotImplemented

    # Method: __mul__
    def __mul__(self,other):

        return self.__mulhandle__(other)

    # Method: __rmul__
    def __rmul__(self,other):

        return self.__mulhandle__(other)

    # Method: __mulhandle__
    def __mulhandle__(self,other):
        """
        Handles left and right multiply.
        Important: If other is a constant, the multiplication
        tree formed has the constant object as the first child.

        :param other: Other operand.
        """

        # Create operator
        operator = cvxpy_obj(OPERATOR,np.nan,MULTIPLICATION)

        # Number
        if np.isscalar(other):

            # Multiplication by 0
            if other == 0.0:
                return 0.0

            # Multiplication by 1
            elif other == 1.0:
                return self

            # Distribution over addition
            elif (type(self) is cvxpy_tree and
                  self.item.name == SUMMATION):
                new_children = []
                for child in self.children:
                    new_children += [other*child]
                return cvxpy_tree(self.item,new_children)

            # Associativity
            elif (type(self) is cvxpy_tree and
                  self.item.name == MULTIPLICATION and
                  self.children[0].type == CONSTANT):
                new_object = other*self.children[0]
                if new_object.value == 1.0:
                    return self.children[1]
                else:
                    return cvxpy_tree(self.item,[new_object,
                                                 self.children[1]])

            # Constant times constant
            elif type(self) is cvxpy_obj:
                new_const = cvxpy_obj(CONSTANT,self.value*other,
                                      str(self.value*other))
                return new_const

            # Else
            else:
                constant = cvxpy_obj(CONSTANT,other,str(other))
                return cvxpy_tree(operator,[constant,self])

        # Matrix
        elif type(other) is cvxpy_matrix:
            (m,n) = other.shape
            new_ar = cvxpy_array(m,n)
            for i in range(0,m,1):
                for j in range(0,n,1):
                    new_ar[i,j] = other[i,j]*self
            return new_ar

        # Sparse matrix
        elif type(other) is cvxpy_spmatrix:
            (m,n) = other.shape
            new_ar = cvxpy_sparray(m,n)
            for i in range(0,m,1):
                rowi_indeces = other.rows[i]
                rowi_values = other.data[i]
                for k in range(0,len(rowi_indeces)):
                    new_ar[i,rowi_indeces[k]] = rowi_values[k]*self
            return new_ar

        # Scalar param
        elif type(other) is cvxpy_scalar_param:
            return cvxpy_tree(operator,[other,self])

        # Tree
        elif type(other) is cvxpy_tree:
            if not len(self.variables):
                return cvxpy_tree(operator,[self,other])
            elif not len(other.variables):
                return cvxpy_tree(operator,[other,self])
            else:
                return NotImplemented

        # Not implemented
        else:
            return NotImplemented

    # Method: __neg__
    def __neg__(self):

        return (-1.0)*self

    # Method: __str__
    def __str__(self):

        return self.name

#***********************************************************************#
# Class definition: cvxpy_scalar_var                                    #
#***********************************************************************#
class cvxpy_scalar_var(cvxpy_obj):

    # Variable counter
    i = 0

    # Method: __init__
    def __init__(self,name=None):
        """
        Class constructor.

        :param name: String.
        """

        # Assign a name
        if name is None:
            name = BASE_VAR+str(cvxpy_scalar_var.i)
        cvxpy_scalar_var.i += 1

        # Call parent constructor
        cvxpy_obj.__init__(self,VARIABLE,np.nan,name)

    # Method: __getattribute__
    def __getattribute__(self,name):

        # Variables
        if name == 'variables':
            return cvxpy_list([self])

        # Other
        else:
            return object.__getattribute__(self,name)

    # Method: is_nonnegative_constant
    def is_nonnegative_constant(self):
        """
        Determines if object is a nonnegative constant.
        """

        return False

    # Method: is_nonpositive_constant
    def is_nonpositive_constant(self):
        """
        Determines if object is a nonpositive constant.
        """

        return False

#***********************************************************************#
# Class definition: cvxpy_scalar_param                                  #
#***********************************************************************#
class cvxpy_scalar_param(cvxpy_obj):

    # Param counter
    i = 0

    # Method: __init__
    def __init__(self,attribute=None,name=None):
        """
        Class constructor.

        :param name: String.
        :param attribute: Keyword (See cvxpy.defs)
        """

        # Assign a name
        if name is None:
            name = BASE_PARAM+str(cvxpy_scalar_param.i)
        cvxpy_scalar_param.i += 1

        # Call parent constructor
        cvxpy_obj.__init__(self,PARAMETER,np.nan,name)

        # Store attribute
        self.attribute = attribute

    # Method: __getattribute__
    def __getattribute__(self,name):

        # Parameters
        if name == 'parameters':
            return cvxpy_list([self])

        # Other
        else:
            return object.__getattribute__(self,name)

    # Method: __setattr__
    def __setattr__(self,name,value):

        # Value
        if name == 'value':
            if not np.isscalar(value):
                raise TypeError('Invalid value type')
            if value < 0 and self.attribute == NONNEGATIVE:
                raise ValueError('Value must be nonnegative')
            if value > 0 and self.attribute == NONPOSITIVE:
                raise ValueError('Value must be nonpositive')

        # Attribute
        if name == 'attribute':
            if value not in [None,NONNEGATIVE,NONPOSITIVE]:
                raise ValueError("Invalid attribute")
            if value == NONNEGATIVE and self.value < 0:
                self.value = np.nan
            if value == NONPOSITIVE and self.value > 0:
                self.value = np.nan

        # Set attribute
        object.__setattr__(self,name,value)

    # Method: is_nonnegative_constant
    def is_nonnegative_constant(self):
        """
        Determines if object is a nonnegative constant.
        """

        return self.attribute == NONNEGATIVE

    # Method: is_nonpositive_constant
    def is_nonpositive_constant(self):
        """
        Determines if object is a nonpositive constant.
        """

        return self.attribute == NONPOSITIVE

#***********************************************************************#
# Class definition: cvxpy_tree                                          #
#***********************************************************************#
class cvxpy_tree(cvxpy_obj):

    # Method: __init__
    def __init__(self,item,children):
        """
        Class constructor.

        :param item: Operator or function.
        :param children: List of scalar objects.
        """

        self.item = item
        self.children = children
        cvxpy_obj.__init__(self,TREE,np.nan,'')

    # Method: __getattribute__
    def __getattribute__(self,name):

        # Variables
        if name == 'variables':
            l = list(map(lambda x: x.variables,self.children))
            return cvxpy_list(set(reduce(lambda x,y:x+y,l,[])))

        # Parameters
        elif name == 'parameters':
            l = list(map(lambda x: x.parameters,self.children))
            return cvxpy_list(set(reduce(lambda x,y:x+y,l,[])))

        # Value
        elif name == 'value':

            # Summation
            if (self.item.type == OPERATOR and
                self.item.name == SUMMATION):
                return np.sum(list(map(lambda x:x.value,self.children)))

            # Multiplication
            elif (self.item.type == OPERATOR and
                  self.item.name == MULTIPLICATION):
                return self.children[0].value*self.children[1].value

            # Function
            elif self.item.type == FUNCTION:
                return self.item(list(map(lambda x: x.value,self.children)))

            # Error
            else:
                raise TypeError('Invalid tree item')

        # Other
        else:
            return object.__getattribute__(self,name)

    # Method: __str__
    def __str__(self):

        # Multiplication
        if (self.item.type == OPERATOR and
            self.item.name == MULTIPLICATION):
            left_text = str(self.children[0])
            if (self.children[0].type == TREE and
                self.children[0].item.type == OPERATOR):
                left_text = '('+left_text+')'
            right_text = str(self.children[1])
            if (self.children[1].type == TREE and
                self.children[1].item.type == OPERATOR):
                    right_text = '('+right_text+')'
            return left_text+'*'+right_text

        # Summation
        elif (self.item.type == OPERATOR and
              self.item.name == SUMMATION):
            text = str(self.children[0])
            for i in range(1,len(self.children),1):
                x = self.children[i]
                text += ' + '+str(x)
            return text

        # Function
        elif self.item.type == FUNCTION:
            children_text = list(map(str,self.children))
            args = ''
            for i in range(0,len(children_text),1):
                args += children_text[i]
                if i != len(children_text)-1:
                    args += ', '
            return self.item.name+'('+args+')'

        # Error
        else:
            raise TypeError('Invalid tree item')

    # Method: is_convex
    def is_convex(self):
        """
        Determines if tree follows DCP rules of convexity.
        """

        # Summation
        if (self.item.type == OPERATOR and
            self.item.name == SUMMATION):
            return all(list(map(lambda x: x.is_convex(),self.children)))

        # Multiplication
        elif (self.item.type == OPERATOR and
              self.item.name == MULTIPLICATION):
            ob1 = self.children[0]
            ob2 = self.children[1]
            case1 = ob1.is_nonnegative_constant() and ob2.is_convex()
            case2 = ob1.is_nonpositive_constant() and ob2.is_concave()
            case3 = ob2.is_affine()
            return case1 or case2 or case3

        # Function
        elif self.item.type == FUNCTION:
            fcn = self.item
            args = self.children
            return (fcn.action == MINIMIZE) and fcn.is_dcp(args)

        # Error
        else:
            raise TypeError('Invalid tree item')

    # Method: is_concave
    def is_concave(self):
        """
        Determines if tree follows DCP rules of concavity.
        """

        # Summation
        if (self.item.type == OPERATOR and
            self.item.name == SUMMATION):
            return all(list(map(lambda x: x.is_concave(),self.children)))

        # Multiplication
        elif (self.item.type == OPERATOR and
              self.item.name == MULTIPLICATION):
            ob1 = self.children[0]
            ob2 = self.children[1]
            case1 = ob1.is_nonnegative_constant() and ob2.is_concave()
            case2 = ob1.is_nonpositive_constant() and ob2.is_convex()
            case3 = ob2.is_affine()
            return case1 or case2 or case3

        # Function
        elif self.item.type == FUNCTION:
            fcn = self.item
            args = self.children
            return (fcn.action == MAXIMIZE) and fcn.is_dcp(args)

        # Error
        else:
            raise TypeError('Invalid tree item')

    # Method: is_dcp
    def is_dcp(self):
        """
        Determines if tree is DCP-compliant.
        """

        return self.is_convex() or self.is_concave()

    # Method: is_affine
    def is_affine(self):
        """
        Determines if tree is an affine expression.
        """

        # Summation
        if (self.item.type == OPERATOR and
            self.item.name == SUMMATION):
            return all(list(map(lambda x: x.is_affine(),self.children)))

        # Multiplication
        elif (self.item.type == OPERATOR and
              self.item.name == MULTIPLICATION):
            ob2 = self.children[1]
            return ob2.is_affine()

        # Function
        elif self.item.type == FUNCTION:
            return len(cvxpy_list(self.children).variables) == 0

        # Error
        else:
            raise TypeError('Invalid tree item')

    # Method: is_nonnegative_constant
    def is_nonnegative_constant(self):
        """
        Determines if tree is guaranteed to be a
        nonnegative constant.
        """

        # Summation
        if (self.item.type == OPERATOR and
            self.item.name == SUMMATION):
            return all(list(map(lambda x: x.is_nonnegative_constant(),
                           self.children)))

        # Multiplication
        elif (self.item.type == OPERATOR and
              self.item.name == MULTIPLICATION):
            ob1 = self.children[0]
            ob2 = self.children[1]
            both_nonneg = all(list(map(lambda x:x.is_nonnegative_constant(),
                                  [ob1,ob2])))
            both_nonpos = all(list(map(lambda x:x.is_nonpositive_constant(),
                                  [ob1,ob2])))
            return both_nonneg or both_nonpos

        # Function
        elif self.item.type == FUNCTION:
            return False

        # Error
        else:
            raise TypeError('Invalid tree item')

    # Method: is_nonpositive_constant
    def is_nonpositive_constant(self):
        """
        Determines if tree is guaranteed to be a
        nonpositive constant.
        """

        # Summation
        if (self.item.type == OPERATOR and
            self.item.name == SUMMATION):
            return all(list(map(lambda x: x.is_nonpositive_constant(),
                           self.children)))

        # Multiplication
        elif (self.item.type == OPERATOR and
              self.item.name == MULTIPLICATION):
            ob1 = self.children[0]
            ob2 = self.children[1]
            p1 =(ob1.is_nonnegative_constant(),
                 ob1.is_nonpositive_constant())
            p2 =(ob2.is_nonnegative_constant(),
                 ob2.is_nonpositive_constant())
            return (p1[0] and p2[1]) or (p1[1] and p2[0])

        # Function
        elif self.item.type == FUNCTION:
            return False

        # Error
        else:
            raise TypeError('Invalid tree item')

# Load modules
from .arrays import cvxpy_array,cvxpy_sparray
from .arrays import cvxpy_matrix,cvxpy_spmatrix
from .constraints import cvxpy_list



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from ...ir import impulse_response
import scipy.signal as signal

__all__ = ["quantization_noise_gain_by_conv"]


def quantization_noise_gain_by_conv(NTF, H, H_type='zpk', db=80):
    """
    Legacy version of :func:`pydsm.NTFdesign.quantization_noise_gain`

    Computes the quantization noise power gain, based on a convolution

    Parameters
    ----------
    NTF : tuple
        NTF definition in zpk or nd form
    H : tuple or array_like
        output filter definition in zpk or nd form if H_type='zpk' or 'ba'
        (in this case, H is a tuple with 3 or 2 entries);
        output filter impulse response if H_type='imp' (in this case, H is an
        array)
    H_type : str
        type of specification for parameter H. One of: 'zpk', 'ba', or
        'imp'
    db : real
        a precision hint for the computation of impulse responses

    Returns
    -------
    a : real
        noise power gain

    Notes
    -----
    The computation is practiced as the sum of the squared entries
    in the impulse response of the cascaded filter NTF*H
    """
    h1_ir = impulse_response(NTF, db=db)
    if H_type == 'zpk' or H_type == 'ba':
        h2_ir = impulse_response(H, db=db)
    elif H_type == 'imp':
        h2_ir = H
    else:
        raise ValueError("Incorrect filter type specification")
    conv = signal.convolve(h1_ir, h2_ir)
    return np.sum(conv**2)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from .defs import *
from .arrays import *
from .scalars import *
from .constraints import *

# Names
__all__ = ["variable","variable_reset","parameter",
           "parameter_reset", "matrix","spmatrix","equals","eq",
           "less_equals","leq","greater_equals","geq","program",
           "minimize","maximize","belongs"]

# Interface function: variable
def variable(m=1,n=1,structure=None,name=None):
    """
    | Creates an optimization variable.

    :param m: rows.
    :param n: columns.
    :param structure: ``None``,
                      ``'symmetric'``,
                      ``'lower_triangular'`` or
                      ``'upper_triangular'``.
    :param name: string or
                 ``None``.
    :return: :ref:`scalar variable<scalar_var_obj>` or
             :ref:`multidimensional variable<multi_var_obj>`.
    """

    if (m,n) == (1,1):
        return cvxpy_scalar_var(name)
    else:
        return cvxpy_var(m,n,structure,name)

# Interface function: variable_reset
def variable_reset():
    """
    | Resets variable counter.
    """

    cvxpy_scalar_var.i = 0

# Interface function: parameter
def parameter(m=1,n=1,attribute=None,name=None):
    """
    | Creates a parameter.

    :param m: rows.
    :param n: columns.
    :param attribute: ``None``,
                      ``'nonnegative'`` or
                      ``'nonpositive'``.
    :param name: string or
                 ``None``.
    :return: :ref:`scalar parameter<scalar_param_obj>` or
             :ref:`multidimensional parameter<multi_param_obj>`.
    """

    if (m,n) == (1,1):
        return cvxpy_scalar_param(attribute,name)
    else:
        return cvxpy_param(m,n,attribute,name)

# Interface function: parameter_reset
def parameter_reset():
    """
    | Resets parameter counter.
    """

    cvxpy_scalar_param.i = 0

# Interface function: matrix
def matrix(data):
    """
    | Creates a matrix. See **numpy.matrix**.

    :param data: array-like or string.
    :return: :ref:`matrix<matrix_obj>`.
    """

    # Check if data is numpy matrix
    if type(data) is np.matrix:
        data = np.array(data)

    # Return matrix
    return cvxpy_matrix(data,np.float64)

# Interface function: spmatrix
def spmatrix(data):
    """
    | Creates a sparse matrix. See **scipy.sparse.lil_matrix**.

    :param data: Dense matrix, sparse matrix or shape tuple.
    :return: :ref:`sparse matrix<spmatrix_obj>`.
    """

    # Return sparse matrix
    return cvxpy_spmatrix(data,np.float64)

# Interface function: equals
def equals(l,r):
    r"""
    | Forms constraint :math:`l = r`.

    :param l: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :param r: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: :ref:`constraint<constr_obj>` or
             :ref:`list of constraints<constr_list_obj>`.
    """
    return compare(l,EQUALS,r)

# Interface function: eq
eq = equals

# Interface function: less_equals
def less_equals(l,r):
    r"""
    | Forms constraint :math:`l \leq r`.

    :param l: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :param r: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: :ref:`constraint<constr_obj>` or
             :ref:`list of constraints<constr_list_obj>`.
    """
    return compare(l,LESS_EQUALS,r)

# Interface function: leq
leq = less_equals

# Interface function: greater_equals
def greater_equals(l,r):
    r"""
    | Forms constraint :math:`l \geq r`.

    :param l: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :param r: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: :ref:`constraint<constr_obj>` or
             :ref:`list of constraints<constr_list_obj>`.
    """
    return compare(l,GREATER_EQUALS,r)

# Interface function: geq
geq = greater_equals

# Interface function: program
def program(pair,constraints=[],formals=[],options=None,name='prog'):
    """
    | Creates an optimization program.

    :param pair: output of function
                 :func:`minimize() <cvxpy.interface.minimize>` or
                 :func:`maximize() <cvxpy.interface.maximize>`.
    :param constraints: list of :ref:`constraints<constr_obj>`.
    :param formals: list of
                    :ref:`scalar variables<scalar_var_obj>`,
		    :ref:`scalar parameters<scalar_param_obj>`,
                    :ref:`multidimensional variables<multi_var_obj>` or
		    :ref:`multidimensional parameters<multi_param_obj>`.
    :param options: dictionary or
                    ``None``.
    :param name: string.
    :return: :ref:`program<program_obj>`.
    """

    # Get action and objective function
    action = pair[0]
    objective = pair[1]

    # Create and return program
    return cvxpy_program(action,objective,constraints,formals,
                         options,name)

# Interface function: minimize
def minimize(objective):
    """
    | Forms pair (action, objective function).

    :param objective: number or
                      :ref:`scalar object<scalar_ref>`.
    :return: tuple.
    """
    return MINIMIZE,objective

# Interface function: maximize
def maximize(objective):
    """
    | Forms pair (action, objective function).

    :param objective: number or
                      :ref:`scalar object<scalar_ref>`.
    :return: tuple.
    """
    return MAXIMIZE,objective

# Method: belongs
def belongs(x,S):
    r"""
    | Forms constraint :math:`x \in S`.

    :param x: :ref:`multidimensional object<multi_ref>`.
    :param S: set instance.
    :rtype: :ref:`constraint<constr_obj>`.
    """

    # Check first argument
    if ((type(x) is not cvxpy_matrix) and
        (type(x).__name__ not in ARRAY_OBJS)):
        raise TypeError('Invalid first argument')

    # Convert cvxpy_matrix to array
    if type(x) is cvxpy_matrix:
        new_ar = cvxpy_array(x.shape[0],x.shape[1])
        for i in range(0,x.shape[0],1):
            for j in range(0,x.shape[1],1):
                new_ar[i,j] = x[i,j]
        x = new_ar

    # Check second argument
    try:
        if S.type != SET:
            raise TypeError('Invalid second argument')
    except AttributeError:
        raise TypeError('Invalid second argument')

    # Verify dimensions
    if not S.valid_shape(x.shape):
        raise ValueError('Invalid dimensions')

    # Construct and return constraint
    return cvxpy_constr(x,BELONGS,S)

# Load modules
from .programs import *



# -*- coding: utf-8 -*-

# Copyright (c) 2014–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

u"""
Helper functions for NTF design (:mod:`pydsm.NTFdesign.helpers`)
================================================================

This modules provides helper functions for the design of the NTF
of ΔΣ modulators.

.. currentmodule:: pydsm.NTFdesign.helpers


Functions
---------

.. autosummary::
   :toctree: generated/

   maxflat_fir_zeros     -- Zeros of maxflat FIR transfer function
   spread_fir_uc_zeros   -- Zeros spread on unit circle according to cost
"""


from __future__ import division

import numpy as np
from scipy.optimize import minimize
from ..utilities import digested_options

__all__ = ["maxflat_fir_zeros", "spread_fir_uc_zeros"]


def maxflat_fir_zeros(order, alpha):
    """
    Compute the zeros of a maxflat FIR transfer function.

    The computed FIR transfer function is used in the DELSIG
    :func:`pydsm.delsig.synthesizeNTF()` design method
    (also known as :func:`pydsm.NTFdesign.ntf_schreier()`)
    for the denominator of the noise transfer function.

    Parameters
    ----------
    order : int
        the transfer function order
    alpha : float
        parameter controlling the steepness of the magnitude
        response

    Returns
    -------
    zeros : ndarray
        an array of complex roots of the FIR transfer function
    """
    x = 1./np.sqrt(alpha)
    me2 = -0.5*(x**(2./order))
    w = (2*np.arange(1, order+1)+1)*np.pi/order
    mb2 = 1+me2*np.exp(1j*w)
    p = mb2 - np.sqrt(mb2**2-1)
    # Reflect poles that fall out of the unit circle
    out = (np.abs(p) > 1)
    p[out] = 1/p[out]
    return p


def spread_fir_uc_zeros(order, OSR, cf, cf_args=[], cf_kwargs={}, **options):
    """
    Compute the best spreading of zerors on the unit circle.

    The computed FIR transfer function is optimal according to a
    criterion expressed by the cost function ``cf``.

    The computed FIR transfer function is used in the DELSIG
    :func:`pydsm.delsig.synthesizeNTF()` design method
    (also known as :func:`pydsm.NTFdesign.ntf_schreier()`)
    for the numerator of the noise transfer function.

    Parameters
    ----------
    order : int
        the transfer function order
    OSR : float
        the oversampling ratio
    cf : function
        cost function for the optimization. Takes a transfer
        function in zpk form as the first argument plus more
        arguments as required
    cf_args: list
        positional args of function ``cf``
    cf_kwargs: dict
        keyword args of function ``cf``

    Returns
    -------
    zeros : ndarray
        an array of complex roots of the FIR transfer function

    Other parameters
    ----------------
    L-BFGS-B_opts : dictionary, optional
        Parameters passed to the ``F-BFGS-B``
        optimizer. Allowed options are:

        ``ftol``
            stop condition for the minimization
        ``gtol``
            gradient stop condition for the minimization
        ``maxcor``
            max number of variables used in hessian approximation
        ``maxiter``
            max number of iterations
        ``maxfun``
            max number of function evaluations
        ``eps``
            Step size used for numerical approximation of the jacobian

        Do not use other options since they could break the minimizer in
        unexpected ways. Defaults can be set by changing the function
        ``default_options`` attribute.

    Notes
    -----
    The system is implicitly assumed to be low-pass. Hence, the zeros
    are spread on the unit circle in the [0, pi/OSR] range.

    See Also
    --------
    scipy.optimize.minimize :  for the parameters passed to the ``L-BFGS-B``
        minimizer
    """
    def dof2zeros(xx):
        zeros[0:xl] = np.exp(1j*xx)
        zeros[xl:2*xl] = zeros[0:xl].conj()
        if order % 2 != 0:
            zeros[-1] = 1
        return zeros

    def mf(xx):
        return np.log10(cf((dof2zeros(xx), np.zeros(order), 1),
                           *cf_args, **cf_kwargs))

    # Manage optional parameters
    opts = digested_options(options, spread_fir_uc_zeros.default_options,
                            [], ['L_BFGS_B_opts'])
    xl = order // 2
    zeros = np.zeros(order, dtype=complex)
    xx = minimize(mf, np.linspace(np.pi/OSR/order, np.pi/OSR, xl),
                  method='l-bfgs-b', options=opts["L_BFGS_B_opts"],
                  bounds=[(0, np.pi/OSR)] * xl).x
    return dof2zeros(xx)

spread_fir_uc_zeros.default_options = {"L_BFGS_B_opts":
                                       {"ftol": 2.220446049250313e-09,
                                        "gtol": 1e-05,
                                        "maxcor": 10,
                                        "maxiter": 15000,
                                        "maxfun": 15000,
                                        "eps": 1E-8}}



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.


"""
Delsig based NTF synthesis (:mod:`pydsm.NTFdesign.delsig`)
==========================================================

This modules provides code for the synthesis of the modulator NTF, based
on the routines in the DELSIG toolkit by R. Schreier.

These routines are also available in the :mod:`pydsm.delsig` and get
abbreviated names here.

.. currentmodule:: pydsm.NTFdesign.delsig


Functions
---------

.. autosummary::
   :toctree: generated/

    ntf_schreier    -- Design NTF with Schreier's method
    ntf_chebyshev   -- Design NTF based on Chebyshev type II form
    ntf_clans       -- Design NTF based on CLANS mehtod


Deprecated functions
--------------------

.. autosummary::
   :toctree: generated/

   synthesizeNTF           -- Alias for `ntf_schreier`
   synthesizeChebyshevNTF  -- Alias for `ntf_chebyshev`
   clans                   -- Alias for `ntf_clans`
"""

from __future__ import division, print_function

from ..delsig import (synthesizeNTF as ntf_schreier,
                      synthesizeChebyshevNTF as ntf_chebyshev,
                      clans as ntf_clans)
from warnings import warn
from ..exceptions import PyDsmDeprecationWarning

__all__ = ['ntf_schreier', 'ntf_chebyshev', 'ntf_clans',
           'synthesizeNTF', 'synthesizeChebyshevNTF', 'clans']


# Following part is deprecated

def synthesizeNTF(order=3, osr=64, opt=0, H_inf=1.5, f0=0.0,
                  **options):
    """
    Alias of :func:`ntf_schreier`

    .. deprecated:: 0.11.0
        Function has been moved to the :mod:`pydsm.NTFdesign` module with
        name :func:`ntf_schreier`.
    """
    warn("Function superseded by ntf_schreier in "
         "NTFdesign module", PyDsmDeprecationWarning)
    return ntf_schreier(order, osr, opt, H_inf, f0, **options)


def synthesizeChebyshevNTF(order=3, osr=64, opt=0, H_inf=1.5, f0=0.0):
    """
    Alias of :func:`ntf_chebyshev`

    .. deprecated:: 0.11.0
        Function has been moved to the :mod:`pydsm.NTFdesign` module with
        name :func:`ntf_chebyshev`.
    """
    warn("Function superseded by ntf_chebyshev in "
         "NTFdesign module", PyDsmDeprecationWarning)
    return ntf_chebyshev(order, osr, opt, H_inf, f0)


def clans(order=4, osr=64, nq=5, rmax=0.95, opt=0, **options):
    """
    Alias of :func:`ntf_clans`

    .. deprecated:: 0.11.0
        Function has been moved to the :mod:`pydsm.NTFdesign` module with
        name :func:`ntf_clans`.
    """
    warn("Function superseded by ntf_clans in "
         "NTFdesign module", PyDsmDeprecationWarning)
    return ntf_clans(order, osr, nq, rmax, opt, **options)



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
DELSIG style matrix and vector padding routines
===============================================
"""

import numpy as np

__all__ = ["padl", "padr", "padt", "padb"]


def padl(x, n, val=0):
    """
    Pad a matrix or a vector on the left.

    Parameters
    ----------
    x : ndarray
        matrix or vector to be padded (must be 1D or 2D)
    n : int
        (column) length to pad x to
    val : scalar, optional
        fill value for padding. Defaults to 0.

    Returns
    -------
    y : ndarray
        matrix or vector x, padded to the left to have n entries (if 1D)
        or n columns (if 2D). Padding value is val.

    Notes
    -----
    If x is an array rather than a matrix, then an array is returned,
    padded at its beginning. Namely, the array is interpreted as a row
    vector.
    """
    if x.ndim == 1:
        return np.hstack((val*np.ones(n-x.shape[0]), x))
    return np.hstack((val*np.ones((x.shape[0], n-x.shape[1])), x))


def padr(x, n, val=0):
    """Pad a matrix or a vector x on the right.

    Parameters
    ----------
    x : ndarray
        matrix or vector to be padded (must be 1D or 2D)
    n : int
        (column) length to pad x to
    val : scalar, optional
        fill value for padding. Defaults to 0.

    Returns
    -------
    y : ndarray
        matrix or vector x, padded to the right to have n entries (if 1D)
        or n columns (if 2D). Padding value is val.

    Notes
    -----
    If x is an array rather than a matrix, then an array is returned,
    padded at its end. Namely, the array is interpreted as a row
    vector.
    """
    if x.ndim == 1:
        return np.hstack((x, (val*np.ones(n-x.shape[0]))))
    return np.hstack((x, (val*np.ones((x.shape[0], n-x.shape[1])))))


def padt(x, n, val=0):
    """Pad a matrix or a vector on the top.

    Parameters
    ----------
    x : ndarray
        matrix or vector to be padded (must be 1D or 2D)
    n : int
        row height to pad x to
    val : number, optional
        fill value for padding. Defaults to 0.

    Returns
    -------
    y : ndarray
        matrix or vector x, padded to the top to have n rows.
        Padding value is val.

    Notes
    -----
    If x is an array rather than a matrix, then an array is returned,
    padded at its beginning. Namely, the array is interpreted as a column
    vector.
    """
    if x.ndim == 1:
        return np.hstack((val*np.ones(n-x.shape[0]), x))
    return np.vstack((val*np.ones((n-x.shape[0], x.shape[1])), x))


def padb(x, n, val=0):
    """Pad a matrix or a vector on the bottom.

    Parameters
    ----------
    x : ndarray
        matrix or vector to be padded (must be 1D or 2D)
    n : int
        row height to pad x to
    val : scalar, optional
        fill value for padding. Defaults to 0.

    Returns
    -------
    y : ndarray
        matrix or vector x, padded to the bottom to have n rows.
        Padding value is val.

    Notes
    -----
    If x is an array rather than a matrix, then an array is returned,
    padded at its end. Namely, the array is interpreted as a column
    vector.
    """
    if x.ndim == 1:
        return np.hstack((x, (val*np.ones(n-x.shape[0]))))
    return np.vstack((x, val*np.ones((n-x.shape[0], x.shape[1]))))



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from pydsm.NTFdesign import ntf_hybrid_weighting
import pytest

__all__ = ["TestNTF_Hybrid"]


class TestNTF_Hybrid:

    @classmethod
    def setup_class(cls):
        # This test emulates a Schreier-type design using the hybrid
        # design method

        # Set the main design parameters
        cls.order = 3
        cls.OSR = 64

        # Set the NTF z, p, k that would be returned by Scheier's method
        cls.e_k = 1
        e_z = [1.0000, 0.9993 - 0.0382j, 0.9993 + 0.0382j]
        cls.e_z = np.sort(e_z)
        e_p = [0.6692, 0.7652 - 0.2795j, 0.7652 + 0.2795j]
        cls.e_p = np.sort(e_p)

    # Prepare the weighting function for the hybrid method
    def w(self, f):
        return 1. if f <= 0.5/self.OSR else 1E-12

    def test_ntf_hybrid_cvxpy_tdr(self):
        try:
            from pydsm import cvxpy_tdr     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy_old' not installed")
        z, p, k = ntf_hybrid_weighting(self.order, self.w, H_inf=1.5,
                                       poles=self.e_p,
                                       show_progress=False,
                                       modeler='cvxpy_old',
                                       quad_opts={"points": [0.5/self.OSR]},
                                       cvxopt_opts={"reltol": 1E-14,
                                                    "abstol": 1E-15})
        z = np.sort(z)
        p = np.sort(p)
        np.testing.assert_allclose(k, self.e_k, 1e-6)
        np.testing.assert_allclose(z, self.e_z, 3e-4)
        np.testing.assert_allclose(p, self.e_p, 3e-4)

    def test_ntf_hybrid_cvxpy_cvxopt(self):
        try:
            import cvxpy     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy' not installed")
        z, p, k = ntf_hybrid_weighting(self.order, self.w, H_inf=1.5,
                                       poles=self.e_p,
                                       show_progress=False,
                                       modeler='cvxpy',
                                       quad_opts={"points": [0.5/self.OSR]},
                                       cvxopt_opts={"reltol": 1E-14,
                                                    "abstol": 2E-16})
        z = np.sort(z)
        p = np.sort(p)
        np.testing.assert_allclose(k, self.e_k, 1e-6)
        np.testing.assert_allclose(z, self.e_z, 3e-4)
        np.testing.assert_allclose(p, self.e_p, 3e-4)

    def test_ntf_hybrid_cvxpy_scs(self):
        try:
            import cvxpy     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy' not installed")
        z, p, k = ntf_hybrid_weighting(self.order, self.w, H_inf=1.5,
                                       poles=self.e_p,
                                       show_progress=False,
                                       modeler='cvxpy',
                                       quad_opts={"points": [0.5/self.OSR]},
                                       cvxpy_opts={"solver": "scs"},
                                       scs_opts={"eps": 1E-10,
                                                 "max_iters": 100000,
                                                 "acceleration_lookback": 10})
        z = np.sort(z)
        p = np.sort(p)
        np.testing.assert_allclose(k, self.e_k, 1e-6)
        np.testing.assert_allclose(z, self.e_z, 5e-2)
        np.testing.assert_allclose(p, self.e_p, 3e-6)

    def test_ntf_hybrid_picos(self):
        try:
            import picos     # analysis:ignore
        except:
            pytest.skip("Modeler 'picos' not installed")
        z, p, k = ntf_hybrid_weighting(self.order, self.w, H_inf=1.5,
                                       poles=self.e_p,
                                       show_progress=False,
                                       modeler='picos',
                                       quad_opts={"points": [0.5/self.OSR],
                                                  "epsrel": 1E-12},
                                       cvxopt_opts={"reltol": 1E-14,
                                                    "abstol": 1E-15})
        z = np.sort(z)
        p = np.sort(p)
        np.testing.assert_allclose(k, self.e_k, 1e-6)
        np.testing.assert_allclose(z, self.e_z, 3e-4)
        np.testing.assert_allclose(p, self.e_p, 3e-4)



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Compatibility functions for PyDSM (:mod:`pydsm.relab`)
======================================================

This module re-implements some MATLAB interfaces that are useful
for PyDSM.

.. currentmodule:: pydsm.relab


Functions
---------

.. autosummary::
   :toctree: generated/

   eps       -- Floating point relative accuracy
   db        -- Converts a value to dB a la Matlab
   cplxpair  -- Sorts values in input list by complex pairs
   shiftdim  -- Shift dimensions a la Matlab
"""

from __future__ import division, print_function

import numpy as np

__all__ = ["eps", "db", "cplxpair", "shiftdim"]


def eps(x=1.):
    """Provide floating point relative accuracy

    This function tries to replicate the ``eps`` interface of Matlab.

    Parameters
    ----------
    x : float or array_like or numeric type

    Returns
    -------
    d : float or array_like
        If x is a float, the positive distance from abs(x) to the next larger
        in magnitude floating point number of the same precision as x.

        If x is an array, operation is elementwise.

        If x is a numeric type, operation is for 1.0 in that numeric type.

    Examples
    --------
    >>> print(eps())
    2.220446049250313e-16
    >>> print(eps(1/2))
    1.1102230246251565e-16
    >>> print(eps(np.float32))
    1.1920929e-07
    """

    def _eps(xi):
        return np.finfo(xi).eps * np.abs(xi)
    if isinstance(x, (type, np.dtype)):
        return np.finfo(x).eps
    elif np.isscalar(x):
        return _eps(x)
    x = np.asarray(x)
    d = np.empty_like(x)
    d.flat = [_eps(xi) for xi in x.flat]
    return d


def db(x, signal_type='voltage', R=1):
    """
    Converts a value to dB a la Matlab

    This function tries to replicate the ``dB`` interface of Matlab.

    Parameters
    ----------
    x : real
        value to be converted. Should be positive and non null.
    signal_type : string, optional
        either 'voltage' or 'power'. Defaults to 'voltage'
    R : real, optional
        load resistance value. Defaults to 1.

    Returns
    -------
    y : real
        value in dB corresponding to x
        if signal_type is 'power' the result is 10*log10(x). Otherwise (if
        signal_type is 'voltage'), then power is measured over resistor R

    Notes
    -----
    The default R value assures that when signal_type is 'voltage' dB defaults
    to the classical 20*log10(x) computation.
    """
    if signal_type == 'power':
        return 10*np.log10(x)
    else:
        return 10*np.log10(np.abs(x)**2./R)


def shiftdim(x, n=None, nargout=2):
    """
    Shift dimensions a la Matlab

    When n is provided, shiftdim shifts the axes of x by n.
    If n is positive, it shifts the axes to the left, wrapping the
    leading axes with non unitary length to the end.
    When n is negative, it shifts the axes to the right, inserting n leading
    axes with unitary length.
    When n is not provided or None, it shifts the axes to the left, reducing
    the number of dimensions and removing all the leading axes with unitary
    length.

    Parameters
    ----------
    x : array like
        multi-dimensional array to operate upon
    n : int or None, optional
        amount to shift. Defaults to None, which means automatic computation
    nargout : int
        number of output values

    Returns
    -------
    y : ndarray
        the result of the axes shift operation
    n : int
        the actual shift

    Examples
    --------
    >>> from numpy.random import rand
    >>> a = rand(1, 1, 3, 1, 2)
    >>> b, n = shiftdim(a)
    >>> b.shape
    (3, 1, 2)
    >>> n
    2
    >>> c = shiftdim(b, -n, nargout=1)
    >>> np.alltrue(c == a)
    True
    >>> d = shiftdim(a, 3, nargout=1)
    >>> d.shape
    (1, 2, 1, 1, 3)

    >>> b, n = shiftdim([[[1]]])
    >>> b, n
    (array([[[1]]]), 0)
    """
    outsel = slice(nargout) if nargout > 1 else 0
    x = np.asarray(x)
    s = x.shape
    m = next((i for i, v in enumerate(s) if v > 1), 0)
    if n is None:
        n = m
    if n > 0:
        n = n % x.ndim
    if n > 0:
        if n <= m:
            x = x.reshape(s[n:])
        else:
            x = x.transpose(np.roll(range(x.ndim), -n))
    elif n < 0:
            x = x.reshape((1,)*(-n)+x.shape)
    return (x, n)[outsel]


def cplxpair(x, tol=None, dim=None):
    """
    Sorts values into complex pairs a la Matlab.

    The function takes a vector or multidimensional array of of complex
    conjugate pairs or real numbers and rearranges it so that the complex
    numbers are collected into matched pairs of complex conjugates. The pairs
    are ordered by increasing real part, with purely real elements placed
    after all the complex pairs.

    In the search for complex conjugate pairs a relative tolerance equal to
    ``tol`` is used for comparison purposes. The default tolerance is
    100 times the system floating point accuracy.

    If the input vector is a multidimensional array, the rearrangement is done
    working along the axis specifid by the parameter ``dim`` or along the
    first axis with non-unitary length if ``dim`` is not provided.

    Parameters
    ----------
    x : array_like of complex
        x is an array of complex values, with the assumption that it contains
        either real values or complex values in conjugate pairs.
    tol: real, optional
        relative tolerance for the recognition of pairs.
        Defaults to 100 times the system floating point accuracy for the
        specific number type.
    dim: integer, optional
        The axis to operate upon.

    Returns
    -------
    y : ndarray
        y is an array of complex values, with the same values in x, yet now
        sorted as complex pairs by increasing real part. Real elements in x
        are place after the complex pairs, sorted in increasing order.

    Raises
    ------
    ValueError
        'Complex numbers cannot be paired' if there are unpaired complex
        entries in x.

    Examples
    --------
    >>> a = np.exp(2j*np.pi*np.arange(0, 5)/5)
    >>> b1 = cplxpair(a)
    >>> b2 = np.asarray([-0.80901699-0.58778525j, -0.80901699+0.58778525j,
    ...                   0.30901699-0.95105652j,  0.30901699+0.95105652j,
    ...                   1.00000000+0.j])
    >>> np.allclose(b1, b2)
    True

    >>> cplxpair(1)
    array([1])

    >>> cplxpair([[5, 6, 4], [3, 2, 1]])
    array([[3, 2, 1],
           [5, 6, 4]])

    >>> cplxpair([[5, 6, 4], [3, 2, 1]], dim=1)
    array([[4, 5, 6],
           [1, 2, 3]])

    See also
    --------
    eps : the system floating point accuracy
    """

    def cplxpair_vec(x, tol):
        real_mask = np.abs(x.imag) <= tol*np.abs(x)
        x_real = np.sort(np.real(x[real_mask]))
        x_cplx = np.sort(x[np.logical_not(real_mask)])
        if x_cplx.size == 0:
            return x_real
        if (x_cplx.size % 2) != 0:
            raise ValueError('Complex numbers cannot be paired')
        if np.any(np.real(x_cplx[1::2])-np.real(x_cplx[0::2]) >
                  tol*np.abs(x_cplx[0::2])):
            raise ValueError('Complex numbers cannot be paired')
        start = 0
        while start < x_cplx.size:
            sim_len = next((i for i, v in enumerate(x_cplx[start+1:]) if
                           (np.abs(np.real(v)-np.real(x_cplx[start])) >
                            tol*np.abs(v))), x_cplx.size-start-1)+1
            if (sim_len % 2) != 0:
                sim_len -= 1
            # At this point, sim_len elements with identical real part
            # have been identified.
            sub_x = x_cplx[start:start+sim_len]
            srt = np.argsort(np.imag(sub_x))
            sub_x = sub_x[srt]
            if np.any(np.abs(np.imag(sub_x)+np.imag(sub_x[::-1])) >
                      tol*np.abs(sub_x)):
                raise ValueError('Complex numbers cannot be paired')
            # Output should contain "perfect" pairs. Hence, keep entries
            # with positive imaginary parts amd use conjugate for pair
            x_cplx[start:start+sim_len] = np.concatenate(
                (np.conj(sub_x[:sim_len//2-1:-1]),
                 sub_x[:sim_len//2-1:-1]))
            start += sim_len
        return np.concatenate((x_cplx, x_real))

    x = np.atleast_1d(x)
    if x.size == 0:
        return x
    if dim is None:
        dim = next((i for i, v in enumerate(x.shape) if v > 1), 0)
    if tol is None:
        try:
            tol = 100*eps(x.dtype)
        except:
            tol = 100*eps(np.float)
    return np.apply_along_axis(cplxpair_vec, dim, x, tol)
